# -*- coding: utf-8 -*-
"""
Created on Sat Mar  9 16:09:08 2024

@author: dvezinet
"""


import datetime as dtm


import numpy as np
import matplotlib.pyplot as plt
from matplotlib import gridspec
import matplotlib.transforms as mtransforms
import datastock as ds


from . import _class01_plot


# ###############################################################
# ###############################################################
#               Main
# ###############################################################


def main(
    coll=None,
    key=None,
    keyY=None,
    # options
    details=None,
    # uncertainty propagation
    uncertainty_method=None,
    # plotting
    dprop=None,
    vmin=None,
    vmax=None,
    # lines labels
    lines_labels=True,
    lines_labels_color=None,
    lines_labels_rotation=None,
    lines_labels_horizontalalignment=None,
    # figure
    dax=None,
    fs=None,
    dmargin=None,
    tit=None,
    # interactivity
    nmax=None,
    connect=None,
    dinc=None,
    show_commands=None,
    # timing
    timing=None,
):

    # -------------------
    # check
    # -------------------

    (
        key_fit, key_model, key_sol, key_data, key_lamb,
        details, binning,
        lines_labels, dlabels, lines_labels_rotation,
        lines_labels_horizontalalignment,
        connect,
        timing,
    ) = _check(
        coll=coll,
        key=key,
        # options
        details=details,
        # plotting
        lines_labels=lines_labels,
        lines_labels_color=lines_labels_color,
        lines_labels_rotation=lines_labels_rotation,
        lines_labels_horizontalalignment=lines_labels_horizontalalignment,
        # interactivity
        nmax=nmax,
        connect=connect,
        # timing
        timing=timing,
    )

    if timing is True:
        print()
        fname = 'plot_spectral_fit()'
        t1 = dtm.datetime.now()  # DB

    # -------------------
    # interpolate
    # -------------------

    dout = coll.interpolate_spectral_model(
        key_model=key_fit,
        # options
        details=details,
        # uncertainty propagation
        uncertainty_method=uncertainty_method,
        # others
        returnas=dict,
        # timing
        timing=timing,
    )

    if timing is True:
        t2 = dtm.datetime.now()  # DB
        print(f'... timing {fname}: interpolate {(t2-t1).total_seconds()} s')

    # -------------------
    # extract coll2
    # -------------------

    coll2, dkeys, ndim = _extract_coll2(
        coll=coll,
        key_model=key_model,
        key_data=key_data,
        dout=dout,
        details=details,
        keyY=keyY,
    )

    if timing is True:
        t3 = dtm.datetime.now()  # DB
        print(f'... timing {fname}: extract {(t3-t2).total_seconds()} s')

    # -------------------
    # prepare figure
    # -------------------

    if dax is None:
        dax = _get_dax(
            ndim=ndim,# resources
            coll=coll,
            key_data=key_data,
            key_fit=key,
            key_lamb=key_lamb,
            fs=fs,
            dmargin=dmargin,
            tit=tit,
        )

    dax = ds._generic_check._check_dax(dax)

    if timing is True:
        t4 = dtm.datetime.now()  # DB
        print(f'... timing {fname}: get_dax {(t4-t3).total_seconds()} s')

    # -------------------
    # plot
    # -------------------

    if ndim == 1:
        dax = _plot_1d(
            coll2,
            dout=dout,
            dkeys=dkeys,
            dax=dax,
            details=details,
            # lines labels
            dlabels=dlabels,
            lines_labels=lines_labels,
            lines_labels_color=lines_labels_color,
            lines_labels_rotation=lines_labels_rotation,
            lines_labels_horizontalalignment=lines_labels_horizontalalignment,
        )

    elif ndim == 2:
        dax, dgroup = _plot_2d(
            coll=coll,
            key_model=key_model,
            key_fit=key_fit,
            coll2=coll2,
            dout=dout,
            keyY=keyY,
            dkeys=dkeys,
            dax=dax,
            details=details,
            # lines labels
            dlabels=dlabels,
            lines_labels=lines_labels,
            lines_labels_color=lines_labels_color,
            lines_labels_rotation=lines_labels_rotation,
            lines_labels_horizontalalignment=lines_labels_horizontalalignment,
        )

    if timing is True:
        t5 = dtm.datetime.now()  # DB
        print(f'... timing {fname}: plot {(t5-t4).total_seconds()} s')

    # -------------------
    # finalize
    # -------------------

    _finalize_figure(
        dax=dax,
        dout=dout,
        tit=tit,
    )

    if timing is True:
        t6 = dtm.datetime.now()  # DB
        print(f'... timing {fname}: finalize {(t6-t5).total_seconds()} s')

    # ---------------------
    # connect interactivity
    # ---------------------

    if isinstance(dax, dict):
        return dax
    else:
        if connect is True:
            dax.setup_interactivity(kinter='inter0', dgroup=dgroup, dinc=dinc)
            dax.disconnect_old()
            dax.connect()

            dax.show_commands(verb=show_commands)

            if timing is True:
                t7 = dtm.datetime.now()  # DB
                print(f'... timing {fname}: connect {(t7-t6).total_seconds()} s\n')
            return dax
        else:
            return dax, dgroup


# ###############################################################
# ###############################################################
#               check
# ###############################################################


def _check(
    coll=None,
    key=None,
    # options
    details=None,
    # lines labels
    lines_labels=None,
    lines_labels_color=None,
    lines_labels_rotation=None,
    lines_labels_horizontalalignment=None,
    # interactivity
    nmax=None,
    connect=None,
    # timing
    timing=None,
):

    # -------------
    # key
    # -------------

    wsf = coll._which_fit
    lok = [
        k0 for k0, v0 in coll.dobj.get(wsf, {}).items()
        if v0['key_sol'] is not None
    ]

    key = ds._generic_check._check_var(
        key, 'key',
        types=str,
        allowed=lok,
    )

    # key model, data, lamb
    key_model = coll.dobj[wsf][key]['key_model']
    key_sol = coll.dobj[wsf][key]['key_sol']
    key_data = coll.dobj[wsf][key]['key_data']
    key_lamb = coll.dobj[wsf][key]['key_lamb']

    binning = coll.dobj[wsf][key]['dinternal']['binning']

    # -------------
    # details
    # -------------

    details = ds._generic_check._check_var(
        details, 'details',
        types=bool,
        default=True,
    )

    # -------------
    # lines_labels
    # -------------

    if lines_labels in [False, True]:
        labs = None
    else:
        labs = lines_labels

    dlabels = coll.get_spectral_lines_labels(
        keys=key_model,
        labels=labs,
        colors=lines_labels_color,
    )

    # ---------------------
    # lines_labels_rotation
    # ---------------------

    lines_labels_rotation = float(ds._generic_check._check_var(
        lines_labels_rotation, 'lines_labels_rotation',
        types=(float, int),
        default=45,
    ))

    # ---------------------
    # lines_labels_horizontalalignment
    # ---------------------

    lines_labels_horizontalalignment = ds._generic_check._check_var(
        lines_labels_horizontalalignment, 'lines_labels_horizontalalignment',
        types=str,
        default='left',
        allowed=['left', 'center', 'right'],
    )

    # -------------
    # connect
    # -------------

    connect = ds._generic_check._check_var(
        connect, 'connect',
        types=bool,
        default=True,
    )

    # -------------
    # timing
    # -------------

    timing = ds._generic_check._check_var(
        timing, 'timing',
        types=bool,
        default=True,
    )

    return (
        key, key_model, key_sol, key_data, key_lamb,
        details, binning,
        lines_labels, dlabels, lines_labels_rotation,
        lines_labels_horizontalalignment,
        connect,
        timing,
    )


# ###############################################################
# ###############################################################
#               extract coll2
# ###############################################################


def _extract_coll2(
    coll=None,
    key_model=None,
    key_data=None,
    dout=None,
    details=None,
    keyY=None,
):

    coll2, dkeys, ndim = _class01_plot._extract_coll2(
        coll=coll,
        key_model=key_model,
        dout=dout,
        details=details,
        keyY=keyY,
    )

    # ----------
    # add data

    lk = ['data', 'units', 'dim', 'quant', 'ref']
    coll2.add_data(
        key=key_data,
        **{k0: coll._ddata[key_data][k0] for k0 in lk}
    )

    dkeys['data'] = key_data

    # ----------
    # add error

    kerr = 'error'
    lk = ['units', 'dim', 'quant']
    coll2.add_data(
        key=kerr,
        data=coll.ddata[key_data]['data'] - coll2.ddata[dkeys['sum']]['data'],
        **{k0: coll._ddata[key_data][k0] for k0 in lk}
    )

    dkeys['error'] = kerr

    return coll2, dkeys, ndim


# ###############################################################
# ###############################################################
#               plot 1d
# ###############################################################


def _plot_1d(
    coll2=None,
    dout=None,
    dkeys=None,
    dax=None,
    details=None,
    # lines labels
    dlabels=None,
    lines_labels=None,
    lines_labels_color=None,
    lines_labels_rotation=None,
    lines_labels_horizontalalignment=None,
):

    # ------------
    # plot spectrum
    # -----------

    axtype = 'spectrum'
    lax = [kax for kax, vax in dax.items() if axtype in vax['type']]
    for kax in lax:
        ax = dax[kax]['handle']

        # plot fit
        ll, = ax.plot(
            coll2.ddata[dkeys['lamb']]['data'],
            coll2.ddata[dkeys['sum']]['data'],
            ls='-',
            marker='None',
            lw=1.,
            color='k',
        )

        # plot data
        ax.plot(
            coll2.ddata[dkeys['lamb']]['data'],
            coll2.ddata[dkeys['data']]['data'],
            ls='None',
            marker='.',
            lw=1.,
            color='k',
        )

        # details
        if details is True:
            for ff in dkeys['details']:
                ax.plot(
                    coll2.ddata[dkeys['lamb']]['data'],
                    coll2.ddata[ff]['data'],
                    ls='-',
                    marker='None',
                    lw=1.,
                )

        # ------------
        # plot uncertainty
        # -----------

        if dkeys.get('sum_min') is not None:
            # plot fit
            ax.fill_between(
                coll2.ddata[dkeys['lamb']]['data'],
                coll2.ddata[dkeys['sum_min']]['data'],
                coll2.ddata[dkeys['sum_max']]['data'],
                ec='None',
                lw=0.,
                fc=ll.get_color(),
                alpha=0.3,
            )

    # ------------
    # plot diff
    # -----------

    axtype = 'diff'
    lax = [kax for kax, vax in dax.items() if axtype in vax['type']]
    for kax in lax:
        ax = dax[kax]['handle']

        # plot fit
        ax.plot(
            coll2.ddata[dkeys['lamb']]['data'],
            coll2.ddata[dkeys['data']]['data'] - coll2.ddata[dkeys['sum']]['data'],
            ls='-',
            marker='None',
            lw=1.,
            color='k',
        )

    # ----------------
    # add lines_labels
    # ----------------

    if lines_labels is not False:
        raise NotImplementedError()

    return dax


# ###############################################################
# ###############################################################
#               plot 2d
# ###############################################################


def _plot_2d(
    coll=None,
    key_model=None,
    key_fit=None,
    coll2=None,
    dout=None,
    dkeys=None,
    keyY=None,
    dax=None,
    details=None,
    # lines labels
    dlabels=None,
    lines_labels=None,
    lines_labels_color=None,
    lines_labels_rotation=None,
    lines_labels_horizontalalignment=None,
):

    # --------------
    # dvminmax
    # --------------

    vmin = np.nanmin(coll2.ddata[dkeys['data']]['data'])
    vmax = np.nanmax(coll2.ddata[dkeys['data']]['data'])

    dvminmax = {
        'data': {'min': vmin, 'max': vmax}
    }

    errmax = np.nanmax(np.abs(coll2.ddata[dkeys['error']]['data']))
    dvminmax_err = {
        'data': {'min': -errmax, 'max': errmax}
    }

    # --------------
    # plot data
    # --------------

    coll2, dgroup0 = coll2.plot_as_array(
        key=dkeys['data'],
        keyX=dkeys['lamb'],
        keyY=keyY,
        dax={k0: dax[k0] for k0 in ['vert', '2d_data', 'spectrum']},
        aspect='auto',
        dvminmax=dvminmax,
        connect=False,
        inplace=True,
    )

    # --------------
    # plot fit
    # --------------

    coll2, dgroup1 = coll2.plot_as_array(
        key=dkeys['sum'],
        keyX=dkeys['lamb'],
        keyY=keyY,
        dax={k0: dax[k0] for k0 in ['vert', '2d_fit', 'spectrum']},
        aspect='auto',
        dvminmax=dvminmax,
        connect=False,
        inplace=True,
    )

    # --------------
    # plot error
    # --------------

    coll2, dgroup0 = coll2.plot_as_array(
        key=dkeys['error'],
        keyX=dkeys['lamb'],
        keyY=keyY,
        dax={k0: dax[k0] for k0 in ['vert', '2d_err', 'error']},
        aspect='auto',
        dvminmax=dvminmax_err,
        cmap=plt.cm.seismic,
        connect=False,
        inplace=True,
    )

    # --------------
    # plot uncertainty
    # --------------



    # -----------------
    # adjust colors
    # -----------------

    nmax = dgroup0['X']['nmax']
    if lines_labels_color == 'sum':
        for k0, v0 in dlabels.items():
            dlabels[k0]['color'] = [v0['color'] for ii in range(nmax)]

    elif lines_labels_color == 'details':
        ccycle = plt.rcParams['axes.prop_cycle'].by_key()['color']
        ii = 0
        for k0, v0 in dlabels.items():
            dlabels[k0]['color'] = [0 for jj in range(nmax)]
            for jj in range(nmax):
                dlabels[k0]['color'][jj] = ccycle[ii%(len(ccycle))]
                ii += 1
    else:
        for k0, v0 in dlabels.items():
            dlabels[k0]['color'] = [v0['color'] for ii in range(nmax)]

    # ----------------------------------
    # needed bot for details and labels
    # ----------------------------------

    reflamb = coll2.ddata[dkeys['lamb']]['ref'][0]
    refs = coll2.ddata[dkeys['sum']]['ref']
    axis = refs.index(reflamb)

    # --------------
    # plot spectrum
    # --------------

    lls = ['-', '--', '-.', ':']
    if details is True:

        lamb = coll2.ddata[dkeys['lamb']]['data']
        wsm = coll._which_model
        lfunc = coll.dobj[wsm][key_model]['keys']

        refs = (refs[axis - 1],)
        nan = np.full(lamb.shape, np.nan)

        kax = 'spectrum'
        ax = dax[kax]['handle']
        for ii, ff in enumerate(lfunc):

            for jj in range(nmax):

                if ff in dlabels.keys():
                    color = dlabels[ff]['color'][jj]
                else:
                    color = None   # bck

                # plot
                ll, = ax.plot(
                    lamb,
                    nan,
                    ls=lls[jj%len(lls)],
                    marker='None',
                    lw=1.,
                    color=color,
                )

                # add mobile
                xydata = 'ydata'
                km = f'{lfunc[ii]}_{jj}'

                coll2.add_mobile(
                    key=km,
                    handle=ll,
                    refs=(refs,),
                    data=(lfunc[ii],),
                    dtype=[xydata],
                    group_vis='Y',  # 'X' <-> 'Y'
                    axes=kax,
                    ind=jj,
                )

    # ----------------
    # add lines_labels
    # ----------------

    if lines_labels is not False:

        _add_labels(
            # resources
            coll=coll,
            coll2=coll2,
            key_fit=key_fit,
            dkeys=dkeys,
            axis=axis,
            nmax=nmax,
            lls=lls,
            # parameters
            kax='spectrum',
            kax_labels='labels',
            dax=dax,
            dlabels=dlabels,
            lines_labels_color=lines_labels_color,
            lines_labels_rotation=lines_labels_rotation,
            lines_labels_horizontalalignment=lines_labels_horizontalalignment,
        )

    return coll2, dgroup0


# ###############################################################
# ###############################################################
#               labels
# ###############################################################


def _add_labels(
    # resources
    coll=None,
    coll2=None,
    key_fit=None,
    dkeys=None,
    axis=None,
    nmax=None,
    lls=None,
    # parameters
    kax=None,
    kax_labels=None,
    dax=None,
    dlabels=None,
    lines_labels_color=None,
    lines_labels_rotation=None,
    lines_labels_horizontalalignment=None,
):

    # -----------------------
    # add ax_labels to coll2

    coll2.add_axes(key=kax_labels, harmonize=False, **dax[kax_labels])
    ax_labels = coll2.dax[kax_labels]['handle']
    ax = coll2.dax[kax]['handle']

    # -------------
    # prepare data

    # extract moments
    dmom = coll.get_spectral_model_moments(key_fit)

    # prepare
    data = coll2.ddata[dkeys['sum']]['data']
    refs = list(coll2.ddata[dkeys['sum']]['ref'])

    # lamb
    lamb = coll2.ddata[dkeys['lamb']]['data']
    sh_lamb = [lamb.size if ii == axis else 1 for ii in range(data.ndim)]
    sh_argmax = [1 if ii == axis else ss for ii, ss in enumerate(data.shape)]
    lambf = lamb.reshape(sh_lamb)

    # custom ref for lines
    rpxy = 'peaks_n2'
    refs[axis] = rpxy
    refm = tuple([rr for ii, rr in enumerate(refs) if ii != axis])
    # add ref
    coll2.add_ref(rpxy, size=2)

    # vmax
    vmax = coll2.dax[kax]['handle'].get_ylim()[1]
    vmaxf = np.full(sh_argmax, vmax)

    # blended transform
    trans = mtransforms.blended_transform_factory(
        ax_labels.transData,
        ax_labels.transAxes,
    )

    # loop on functions / spectral lines
    nan2 = np.r_[np.nan, np.nan]
    for k0, v0 in dlabels.items():

        # argmax
        argmax = dmom[f'{k0}_argmax']['data']
        argmaxf = argmax.reshape(sh_argmax)

        # get ind
        ind = np.abs(argmaxf - lambf).argmin(axis=axis, keepdims=True)

        # dmax
        dmaxf = np.take_along_axis(data, ind, axis=axis)

        peaks_x = np.concatenate((argmaxf, argmaxf), axis=axis)
        peaks_y = np.concatenate((dmaxf, vmaxf), axis=axis)

        # ---------------
        # add data

        # add argmax (for text)
        key_argmax = f'peaks_argmax_{k0}'
        coll2.add_data(
            key=key_argmax,
            data=argmax,
            units=coll2.ddata[dkeys['lamb']]['units'],
            ref=refm,
        )

        # add peaks_x
        keyx = f'peaks_x_{k0}'
        coll2.add_data(
            key=keyx,
            data=peaks_x,
            units=coll2.ddata[dkeys['lamb']]['units'],
            ref=refs,
        )

        keyy = f'peaks_y_{k0}'
        coll2.add_data(
            key=keyy,
            data=peaks_y,
            units=coll2.ddata[dkeys['sum']]['units'],
            ref=refs,
        )

        # ------------------
        # add vertical lines

        for jj in range(nmax):

            # set color
            color = v0['color'][jj]

            # plot
            ll, = ax.plot(
                nan2,
                nan2,
                ls=lls[jj%len(lls)],
                marker='None',
                lw=1.,
                c=color,
            )

            # add mobile
            km = f'peaks_dashed_{k0}_{jj}'
            coll2.add_mobile(
                key=km,
                handle=ll,
                refs=(refm, refm),
                data=(keyx, keyy),
                dtype=['xdata', 'ydata'],
                group_vis='Y',  # 'X' <-> 'Y'
                axes=kax,
                ind=jj,
            )

            # ---------------------
            # add labels above axes

            ll = ax_labels.text(
                0.,
                0.05,
                v0['label'],
                size=12,
                color=ll.get_color(),
                rotation=lines_labels_rotation,
                horizontalalignment=lines_labels_horizontalalignment,
                verticalalignment='bottom',
                transform=trans,
            )

            # add mobile
            km = f'peaks_text_{k0}_{jj}'
            coll2.add_mobile(
                key=km,
                handle=ll,
                refs=(refm,),
                data=(key_argmax,),
                dtype=['x'],
                group_vis='Y',  # 'X' <-> 'Y'
                axes=kax_labels,
                ind=jj,
            )

    return


# ###############################################################
# ###############################################################
#               dax
# ###############################################################


def _get_dax(
    ndim=None,
    # resources
    coll=None,
    key_data=None,
    key_fit=None,
    key_lamb=None,
    # figure
    fs=None,
    dmargin=None,
    tit=None,
):

    if ndim == 1:
        return _get_dax_1d(
            fs=fs,
            dmargin=dmargin,
            tit=tit,
        )

    if ndim == 2:
        return _get_dax_2d(
            coll=coll,
            key_data=key_data,
            key_fit=key_fit,
            key_lamb=key_lamb,
            # options
            fs=fs,
            dmargin=dmargin,
            tit=tit,
        )


def _get_dax_1d(
    fs=None,
    dmargin=None,
    tit=None,
):

    # ---------------
    # check inputs
    # ---------------

    if fs is None:
        fs = (10, 6)

    if dmargin is None:
        dmargin = {
            'left': 0.10, 'right': 0.90,
            'bottom': 0.1, 'top': 0.90,
            'wspace': 0.1, 'hspace': 0.1,
        }

    # ---------------
    # prepare figure
    # ---------------

    dax = {}
    fig = plt.figure(figsize=fs)
    gs = gridspec.GridSpec(3, 1, **dmargin)

    # ------------
    # add axes
    # ------------

    # spectrum
    ax = fig.add_subplot(gs[:2, 0])
    # ax.set_xlabel()
    # ax.set_ylabel()
    dax['1d'] = {'handle': ax, 'type': 'spectrum'}

    ax = fig.add_subplot(gs[2, 0])
    dax['diff'] = {'handle': ax, 'type': 'diff'}

    return dax


def _get_dax_2d(
    coll=None,
    key_data=None,
    key_fit=None,
    key_lamb=None,
    # options
    fs=None,
    dmargin=None,
    tit=None,
):

    # ---------------
    # check inputs
    # ---------------

    if fs is None:
        fs = (19, 10)

    if dmargin is None:
        dmargin = {
            'left': 0.03, 'right': 0.99,
            'bottom': 0.06, 'top': 0.90,
            'wspace': 0.30, 'hspace': 0.30,
        }

    # ---------------
    # prepare figure
    # ---------------

    data_lab = f"{key_data} ({coll.ddata[key_data]['units']})"
    lamb_lab = f"{key_lamb} ({coll.ddata[key_lamb]['units']})"

    # ---------------
    # prepare figure
    # ---------------

    dax = {}
    fig = plt.figure(figsize=fs)
    gs = gridspec.GridSpec(8, 14, **dmargin)

    # ------------
    # add axes
    # ------------

    # --------
    # images

    ax = fig.add_subplot(gs[:, 1:3])
    ax.set_title(f'data\n{key_data}', size=12, fontweight='bold')
    # ax.set_ylabel()
    ax0 = ax
    dax['2d_data'] = {'handle': ax, 'type': 'matrix'}

    ax = fig.add_subplot(gs[:, 3:5], sharex=ax0, sharey=ax0)
    ax.set_title(f'fit\n{key_fit}', size=12, fontweight='bold')
    # ax.set_xlabel()
    # ax.set_ylabel()
    dax['2d_fit'] = {'handle': ax, 'type': 'matrix'}

    ax = fig.add_subplot(gs[:, 5:7], sharex=ax0, sharey=ax0)
    ax.set_title('error', size=12, fontweight='bold')
    # ax.set_xlabel()
    # ax.set_ylabel()
    dax['2d_err'] = {'handle': ax, 'type': 'matrix'}

    # --------
    # vertical

    ax = fig.add_subplot(gs[:, 0], sharey=ax0)
    # ax.set_xlabel()
    # ax.set_ylabel()
    dax['vert'] = {'handle': ax, 'type': 'vertical'}

    # ----------
    # spectrum

    ax = fig.add_subplot(gs[1:-2, 8:], sharex=ax0)
    ax.set_ylabel(data_lab, size=12, fontweight='bold')
    dax['spectrum'] = {'handle': ax, 'type': 'horizontal'}

    # ----------
    # spectrum error

    ax = fig.add_subplot(gs[-2:, 8:], sharex=ax0)
    ax.set_xlabel(lamb_lab, size=12, fontweight='bold')
    ax.set_ylabel('error', size=12, fontweight='bold')
    dax['error'] = {'handle': ax, 'type': 'horizontal'}

    # ----------
    # spectrum labels

    pos_1d = dax['spectrum']['handle'].get_position()
    pos_2d = ax0.get_position()
    y0 = pos_1d.y0 + pos_1d.height
    height = pos_2d.y1 - y0

    ax = fig.add_axes([pos_1d.x0, y0, pos_1d.width, height], sharex=ax0)
    ax.set_ylim(0, 1)
    ax.axis('off')
    dax['labels'] = {'handle': ax, 'type': 'text'}

    return dax



# ###############################################################
# ###############################################################
#               Finalize figure
# ###############################################################


def _finalize_figure(dax=None, dout=None, tit=None):

    # -------------
    # tit
    # -------------

    titdef = (
        f"Spectral model '{dout['key_model']}'\n"
        f"using data '{dout['key_data']}'"
    )
    tit = ds._generic_check._check_var(
        tit, 'tit',
        types=str,
        default=titdef,
    )

    # -------------
    # tit
    # -------------

    if isinstance(dax, dict):
        fig = list(dax.values())[0]['handle'].figure
    else:
        fig = list(dax.dax.values())[0]['handle'].figure

    if tit is not None:
        fig.suptitle(tit, size=12, fontweight='bold')

    return


# -*- coding: utf-8 -*-
"""
Created on Sat Mar  9 16:09:08 2024

@author: dvezinet
"""


import numpy as np
import matplotlib.pyplot as plt
from matplotlib import gridspec
import matplotlib.transforms as mtransforms
import datastock as ds


from . import _class01_plot


# ###############################################################
# ###############################################################
#               Main
# ###############################################################


def main(
    coll=None,
    key_data=None,
    key_lamb=None,
    keyY=None,
    key_lines=None,
    # plotting
    dprop=None,
    dvminmax=None,
    # lines labels
    lines_labels=True,
    lines_labels_color=None,
    lines_labels_rotation=None,
    lines_labels_horizontalalignment=None,
    # figure
    dax=None,
    fs=None,
    dmargin=None,
    tit=None,
    # interactivity
    nmax=None,
    connect=None,
    dinc=None,
    show_commands=None,
):

    # -------------------
    # check
    # -------------------

    (
        key_data, key_lines,
        lines_labels, dlabels, lines_labels_rotation,
        lines_labels_horizontalalignment,
        connect,
    ) = _check(
        coll=coll,
        key_data=key_data,
        key_lines=key_lines,
        # plotting
        lines_labels=lines_labels,
        lines_labels_color=lines_labels_color,
        lines_labels_rotation=lines_labels_rotation,
        lines_labels_horizontalalignment=lines_labels_horizontalalignment,
    )

    # -------------------
    # prepare figure
    # -------------------

    ndim = coll.ddata[key_data]['data'].ndim
    if dax is None:
        dax = _get_dax(
            ndim=ndim,# resources
            coll=coll,
            key_data=key_data,
            key_lamb=key_lamb,
            fs=fs,
            dmargin=dmargin,
            tit=tit,
        )

    dax = ds._generic_check._check_dax(dax)

    # -------------------
    # plot
    # -------------------

    if ndim == 1:
        dax = _plot_1d(
            coll=coll,
            key_data=key_data,
            key_lamb=key_lamb,
            dax=dax,
            dvminmax=dvminmax,
            # lines labels
            dlabels=dlabels,
            lines_labels=lines_labels,
            lines_labels_color=lines_labels_color,
            lines_labels_rotation=lines_labels_rotation,
            lines_labels_horizontalalignment=lines_labels_horizontalalignment,
        )

    elif ndim == 2:
        dax, dgroup = _plot_2d(
            coll=coll,
            key_data=key_data,
            key_lamb=key_lamb,
            keyY=keyY,
            dax=dax,
            dvminmax=dvminmax,
            # lines labels
            dlabels=dlabels,
            lines_labels=lines_labels,
            lines_labels_color=lines_labels_color,
            lines_labels_rotation=lines_labels_rotation,
            lines_labels_horizontalalignment=lines_labels_horizontalalignment,
        )

    # -------------------
    # finalize
    # -------------------

    _finalize_figure(
        dax=dax,
        key_data=key_data,
        tit=tit,
    )

    # ---------------------
    # connect interactivity
    # ---------------------

    if isinstance(dax, dict):
        return dax
    else:
        if connect is True:
            dax.setup_interactivity(kinter='inter0', dgroup=dgroup, dinc=dinc)
            dax.disconnect_old()
            dax.connect()

            dax.show_commands(verb=show_commands)
            return dax
        else:
            return dax, dgroup


# ###############################################################
# ###############################################################
#               check
# ###############################################################


def _check(
    coll=None,
    key_data=None,
    key_lines=None,
    # lines labels
    lines_labels=None,
    lines_labels_color=None,
    lines_labels_rotation=None,
    lines_labels_horizontalalignment=None,
    # interactivity
    nmax=None,
    connect=None,
):

    # -------------
    # key_data
    # -------------

    lok = [
        k0 for k0, v0 in coll.ddata.items()
        if v0['data'].ndim in [1, 2]
    ]
    key_data = ds._generic_check._check_var(
        key_data, 'key_data',
        types=str,
        allowed=lok,
        extra_msg="plot_spectral_data() only implemented for 1d and 2d data!"
    )

    # -------------
    # key lines
    # -------------

    if isinstance(key_lines, str):
        key_lines = [key_lines]

    wsl = coll._which_lines
    lok = list(coll.dobj.get(wsl, {}).keys())
    key_lines = ds._generic_check._check_var_iter(
        key_lines, 'key_lines',
        types=(list, tuple),
        types_iter=str,
        allowed=lok,
        extra_msg="Arg key_lines must be a list of known spectral lines!"
    )

    # -------------
    # lines_labels
    # -------------

    if lines_labels in [False, True]:
        labs = None
    else:
        labs = lines_labels

    dlabels = coll.get_spectral_lines_labels(
        keys=key_lines,
        labels=labs,
        colors=lines_labels_color,
    )

    # ---------------------
    # lines_labels_rotation
    # ---------------------

    lines_labels_rotation = float(ds._generic_check._check_var(
        lines_labels_rotation, 'lines_labels_rotation',
        types=(float, int),
        default=45,
    ))

    # ---------------------
    # lines_labels_horizontalalignment
    # ---------------------

    lines_labels_horizontalalignment = ds._generic_check._check_var(
        lines_labels_horizontalalignment, 'lines_labels_horizontalalignment',
        types=str,
        default='left',
        allowed=['left', 'center', 'right'],
    )

    # -------------
    # connect
    # -------------

    connect = ds._generic_check._check_var(
        connect, 'connect',
        types=bool,
        default=True,
    )

    return (
        key_data, key_lines,
        lines_labels, dlabels, lines_labels_rotation,
        lines_labels_horizontalalignment,
        connect,
    )


# ###############################################################
# ###############################################################
#               plot 1d
# ###############################################################


def _plot_1d(
    coll=None,
    key_data=None,
    key_lamb=None,
    dax=None,
    dvminmax=None,
    # lines labels
    dlabels=None,
    lines_labels=None,
    lines_labels_color=None,
    lines_labels_rotation=None,
    lines_labels_horizontalalignment=None,
):

    # ------------
    # plot spectrum
    # -----------

    axtype = 'spectrum'
    lax = [kax for kax, vax in dax.items() if axtype in vax['type']]
    for kax in lax:
        ax = dax[kax]['handle']

        # plot fit
        ll, = ax.plot(
            coll.ddata[key_lamb]['data'],
            coll.ddata[key_data]['data'],
            ls='-',
            marker='None',
            lw=1.,
            color='k',
        )

    # ----------------
    # add lines_labels
    # ----------------

    if lines_labels is not False:

        _add_labels(
            # resources
            coll=coll,
            # parameters
            ax_spectrum=dax['spectrum']['handle'],
            ax_labels=dax['labels']['handle'],
            dlabels=dlabels,
            lines_labels_rotation=lines_labels_rotation,
            lines_labels_horizontalalignment=lines_labels_horizontalalignment,
        )

    return dax


# ###############################################################
# ###############################################################
#               plot 2d
# ###############################################################


def _plot_2d(
    coll=None,
    key_data=None,
    key_lamb=None,
    keyY=None,
    dax=None,
    dvminmax=None,
    # lines labels
    dlabels=None,
    lines_labels=None,
    lines_labels_color=None,
    lines_labels_rotation=None,
    lines_labels_horizontalalignment=None,
):

    # --------------
    # plot data
    # --------------

    lax = ['2d_data', 'spectrum', 'vert']
    coll2, dgroup0 = coll.plot_as_array(
        key=key_data,
        keyX=key_lamb,
        keyY=keyY,
        dax=dax, # {k0: v0 for k0, v0 in dax.items() if k0 in lax},
        aspect='auto',
        dvminmax=dvminmax,
        connect=False,
        inplace=False,
    )

    # ----------------
    # add lines_labels
    # ----------------

    if lines_labels is not False:

        _add_labels(
            # resources
            coll=coll,
            # parameters
            ax_spectrum=coll2.dax['spectrum']['handle'],
            ax_labels=coll2.dax['labels']['handle'],
            dlabels=dlabels,
            lines_labels_rotation=lines_labels_rotation,
            lines_labels_horizontalalignment=lines_labels_horizontalalignment,
        )

    return coll2, dgroup0


# ###############################################################
# ###############################################################
#               labels
# ###############################################################


def _add_labels(
    # resources
    coll=None,
    # parameters
    ax_spectrum=None,
    ax_labels=None,
    dlabels=None,
    lines_labels_rotation=None,
    lines_labels_horizontalalignment=None,
):

    # -------------
    # prepare data

    # blended transform
    trans = mtransforms.blended_transform_factory(
        ax_labels.transData,
        ax_labels.transAxes,
    )

    # loop on functions / spectral lines
    wsl = coll._which_lines
    for k0, v0 in dlabels.items():

        # argmax
        lamb0 = coll.dobj[wsl][k0]['lamb0']

        # ------------------
        # add vertical lines

        # plot
        ll = ax_spectrum.axvline(
            lamb0,
            ls='--',
            marker='None',
            lw=1.,
            c=v0['color'],
        )

        # ---------------------
        # add labels above axes

        ll = ax_labels.text(
            lamb0,
            0.05,
            v0['label'],
            size=12,
            color=ll.get_color(),
            rotation=lines_labels_rotation,
            horizontalalignment=lines_labels_horizontalalignment,
            verticalalignment='bottom',
            transform=trans,
        )

    return


# ###############################################################
# ###############################################################
#               dax
# ###############################################################


def _get_dax(
    ndim=None,
    # resources
    coll=None,
    key_data=None,
    key_fit=None,
    key_lamb=None,
    # figure
    fs=None,
    dmargin=None,
    tit=None,
):

    if ndim == 1:
        return _get_dax_1d(
            coll=coll,
            key_data=key_data,
            key_lamb=key_lamb,
            fs=fs,
            dmargin=dmargin,
            tit=tit,
        )

    if ndim == 2:
        return _get_dax_2d(
            coll=coll,
            key_data=key_data,
            key_lamb=key_lamb,
            # options
            fs=fs,
            dmargin=dmargin,
            tit=tit,
        )


def _get_dax_1d(
    coll=None,
    key_data=None,
    key_lamb=None,
    fs=None,
    dmargin=None,
    tit=None,
):

    # ---------------
    # check inputs
    # ---------------

    if fs is None:
        fs = (10, 6)

    if dmargin is None:
        dmargin = {
            'left': 0.10, 'right': 0.90,
            'bottom': 0.1, 'top': 0.90,
            'wspace': 0.1, 'hspace': 0.,
        }

    # ---------------
    # prepare labels
    # ---------------

    data_lab = f"{key_data} ({coll.ddata[key_data]['units']})"
    lamb_lab = f"{key_lamb} ({coll.ddata[key_lamb]['units']})"

    # ---------------
    # prepare figure
    # ---------------

    dax = {}
    fig = plt.figure(figsize=fs)
    gs = gridspec.GridSpec(5, 1, **dmargin)

    # ------------
    # add axes
    # ------------

    # spectrum
    ax = fig.add_subplot(gs[1:, 0])
    ax.set_xlabel(lamb_lab, size=12, fontweight='bold')
    ax.set_ylabel(data_lab, size=12, fontweight='bold')
    dax['spectrum'] = {'handle': ax, 'type': 'spectrum'}
    ax0 = ax

    # ----------
    # spectrum labels

    ax = fig.add_subplot(gs[0, 0], sharex=ax0)
    ax.set_ylim(0, 1)
    ax.axis('off')
    dax['labels'] = {'handle': ax, 'type': 'text'}

    return dax


def _get_dax_2d(
    coll=None,
    key_data=None,
    key_lamb=None,
    # options
    fs=None,
    dmargin=None,
    tit=None,
):

    # ---------------
    # check inputs
    # ---------------

    if fs is None:
        fs = (19, 10)

    if dmargin is None:
        dmargin = {
            'left': 0.03, 'right': 0.99,
            'bottom': 0.06, 'top': 0.90,
            'wspace': 0.50, 'hspace': 0.30,
        }

    # ---------------
    # prepare labels
    # ---------------

    data_lab = f"{key_data} ({coll.ddata[key_data]['units']})"
    lamb_lab = f"{key_lamb} ({coll.ddata[key_lamb]['units']})"

    # ---------------
    # prepare figure
    # ---------------

    dax = {}
    fig = plt.figure(figsize=fs)
    gs = gridspec.GridSpec(6, 9, **dmargin)

    # ------------
    # add axes
    # ------------

    # --------
    # images

    ax = fig.add_subplot(gs[:, 1:5])
    ax.set_title(f'data\n{key_data}', size=12, fontweight='bold')
    # ax.set_ylabel()
    ax0 = ax
    dax['2d_data'] = {'handle': ax, 'type': 'matrix'}

    # --------
    # vertical

    ax = fig.add_subplot(gs[:, 0], sharey=ax0)
    # ax.set_xlabel()
    # ax.set_ylabel()
    dax['vert'] = {'handle': ax, 'type': 'vertical'}

    # ----------
    # spectrum

    ax = fig.add_subplot(gs[1:-1, 5:], sharex=ax0)
    ax.set_xlabel(lamb_lab, size=12, fontweight='bold')
    ax.set_ylabel(data_lab, size=12, fontweight='bold')
    dax['spectrum'] = {'handle': ax, 'type': 'horizontal'}

    # ----------
    # spectrum labels

    pos_1d = dax['spectrum']['handle'].get_position()
    pos_2d = ax0.get_position()
    y0 = pos_1d.y0 + pos_1d.height
    height = pos_2d.y1 - y0

    ax = fig.add_axes([pos_1d.x0, y0, pos_1d.width, height], sharex=ax0)
    ax.set_ylim(0, 1)
    ax.axis('off')
    dax['labels'] = {'handle': ax, 'type': 'text'}

    return dax



# ###############################################################
# ###############################################################
#               Finalize figure
# ###############################################################


def _finalize_figure(dax=None, key_data=None, tit=None):

    # -------------
    # tit
    # -------------

    titdef = (
        f"Spectral data '{key_data}'"
    )
    tit = ds._generic_check._check_var(
        tit, 'tit',
        types=str,
        default=titdef,
    )

    # -------------
    # tit
    # -------------

    if isinstance(dax, dict):
        fig = list(dax.values())[0]['handle'].figure
    else:
        fig = list(dax.dax.values())[0]['handle'].figure

    if tit is not None:
        fig.suptitle(tit, size=12, fontweight='bold')

    return


# -*- coding: utf-8 -*-
"""
Created on Thu Aug 22 10:23:58 2024

@author: dvezinet
"""


import numpy as np
import datastock as ds


# #######################################################
# #######################################################
#            Main
# #######################################################


def main(
    coll=None,
    lamb=None,
    binning=None,
    # other fields
    iok=None,
    axis=None,
):

    # ------------------
    # check inputs
    # ------------------

    binning, axis, lamb = _check(coll, binning=binning, lamb=lamb, axis=axis)

    # ------------------
    # trivial
    # ------------------

    if binning is False:
        return False

    # ------------------
    # compute
    # ------------------

    # -----------
    # lamb

    lamb_edges = np.r_[
        lamb[0] - 0.5*(lamb[1] - lamb[0]),
        0.5*(lamb[1:] + lamb[:-1]),
        lamb[-1] + 0.5*(lamb[-1] - lamb[-2]),
    ]

    # in case of non-uniform lamb
    lambd = np.diff(lamb_edges)

    # increments
    lamb_inc = np.linspace(0, 1, binning+1)
    lamb_inc = 0.5*(lamb_inc[1:] + lamb_inc[:-1])
    lamb_inc = lamb_inc[None, :] * lambd[:, None]

    # for later integration
    bin_dlamb = np.repeat(
        lambd / binning,
        binning,
    )

    # get new lamb
    nlamb = lamb.size
    lamb = (lamb_edges[:-1, None] + lamb_inc).ravel()

    # safety check
    assert lamb.size == nlamb * binning

    # --------------
    # update iok_all

    if iok is not None:

        if axis is None:
            if iok.shape == (nlamb,):
                axis = 0
            else:
                msg = "Arg axis must be provided for binning of iok!"
                raise Exception(msg)

        bin_iok = np.repeat(iok, binning, axis=axis)

    else:
        bin_iok = None

    # --------------------------
    # update binning to indices

    ind = np.arange(0, lamb.size, binning)

    # ---------------
    # output
    # ---------------

    dout = {
        'binning': binning,
        'ind': ind,
        'lamb': lamb,
        'dlamb': np.abs(bin_dlamb),
        'iok': bin_iok,
        'edges': lamb_edges,
        'axis': axis,
    }

    return dout


# #######################################################
# #######################################################
#            check
# #######################################################


def _check(
    coll=None,
    binning=None,
    lamb=None,
    axis=None,
):

    # --------------
    # binning
    # --------------

    binning = ds._generic_check._check_var(
        binning, 'binning',
        types=(bool, int),
        default=False,
    )

    # default to 10 if True
    if binning is True:
        binning = 10

    # safety check
    if (binning is not False) and binning <= 0:
        msg = (
            "Arg 'binning' must be a > 0 int\n"
            f"Provided: {binning}"
        )
        raise Exception(msg)

    # ---------------
    # lamb
    # ---------------

    if isinstance(lamb, (list, tuple, np.ndarray)):
        lamb = np.asarray(lamb).ravel()

    elif isinstance(lamb, str):
        lamb = coll.ddata[lamb]['data']

    # -------------
    # axis
    # -------------

    if axis is not None:
        axis = int(ds._generic_check._check_var(
            axis, 'axis',
            types=(int, float),
            sign='>=0',
        ))

    return binning, axis, lamb


"""
This module contains tests for tofu.geom in its structured version
"""


# Built-in
import itertools as itt


# local
from ._setup_teardown import setup_module0, teardown_module0
from .. import nist


#######################################################
#
#     Setup and Teardown
#
#######################################################


def setup_module(module):
    setup_module0(module)


def teardown_module(module):
    teardown_module0(module)


#######################################################
#
#     Creating Ves objects and testing methods
#
#######################################################


class Test_nist(object):

    # ------------------------
    #   setup and teardown
    # ------------------------

    @classmethod
    def setup_class(cls):
        pass

    @classmethod
    def setup_method(self):
        pass

    def teardown_method(self):
        pass

    @classmethod
    def teardown_class(cls):
        pass

    # ------------------------
    #  search online
    # ------------------------

    def test01_search_online_by_wavelengthA(self):

        llambmin = [None, 3.94]
        llambmax = [None, 4.]
        lion = [None, 'H', ['ar', 'W44+']]

        lcache_from = [False, True]
        ldatacol = [False, True]

        # Search by searchstr
        llcomb = [
            llambmin, llambmax, lion,
            lcache_from,
            ldatacol,
        ]
        ii, itot = -1, 2*2*3*2*2
        for comb in itt.product(*llcomb):
            ii += 1
            if all([vv is None for vv in comb[:2]]):
                continue
            if comb[2] == 'H' and comb[1] is None:
                continue
            if comb[2] == 'H' and all([vv is not None for vv in comb[:2]]):
                continue
            if any([vv is None for vv in comb[:2]]) and comb[2] != 'H':
                continue
            print(f'{ii} / {itot}  -  {comb}')

            try:
                out = nist.step01_search_online_by_wavelengthA(
                    lambmin=comb[0],
                    lambmax=comb[1],
                    ion=comb[2],
                    verb=True,
                    return_dout=True,
                    return_dsources=True,
                    cache_from=comb[3],
                    cache_info=True,
                    format_for_DataStock=comb[4],
                    create_custom=True,
                )
                del out

            except Exception as err:

                lstr = [
                    '503 Server Error: Service Unavailable for url:',
                    'File could not be downloaded:',
                    '=> Maybe check internet connection?',
                    # flag that it is running on Github
                    (
                        '/runner/.spectrally/nist/',       # MacOS and linux
                        'runneradmin',    # Windows
                    ),
                ]
                din = {
                    ii: ss in str(err) if isinstance(ss, str)
                    else any([s2 in str(err) for s2 in ss])
                    for ii, ss in enumerate(lstr)
                }
                if all([vv for vv in din.values()]):
                    pass
                else:
                    lstr = [f"\t- {k0}: {v0}" for k0, v0 in din.items()]
                    msg = "\n\n" + "\n".join(lstr) + "\n"
                    raise Exception(msg) from err

    # ------------------------
    #  clear cache
    # ------------------------

    def test02_clear_cache(self):
        nist.clear_cache()


# -*- coding: utf-8 -*-
"""
Created on Thu Feb 15 17:53:54 2024

@author: dvezinet
"""


import copy
import itertools as itt


import numpy as np


#############################################
#############################################
#       MODEL CHECK
#############################################


def _dconstraints(
    coll=None,
    key=None,
    dmodel=None,
    dconstraints=None,
):

    # ----------
    # check
    # ------------

    dconstraints, lvar = _check(
        coll=coll,
        key=key,
        dconstraints=dconstraints,
    )

    # --------------
    # compute matrix
    # --------------

    dconstraints = _compute_coefs_offset(
        dconstraints=dconstraints,
        lvar=lvar,
    )

    # --------------
    # store
    # --------------

    # add ref_nx (number of free parameters)
    knfree = f"nx_{key}"
    if knfree not in coll.dref.keys():
        coll.add_ref(knfree, size=len(dconstraints['groups']))
    elif coll.dref[knfree]['size'] != len(dconstraints['groups']):
        msg = (
            f"add_model('{key}') requires adding ref knfree = '{knfree}'"
            " but it already exists and doesn't have the correct value!\n"
            f"\t- existing: {coll.dref[knfree]['size']}\n"
            f"\t- requested: {len(dconstraints['groups'])}\n"
        )
        raise Exception(msg)

    # dconstraints
    wsm = coll._which_model
    coll._dobj[wsm][key]['ref_nx'] = knfree
    coll._dobj[wsm][key]['dconstraints'] = dconstraints

    return


#############################################
#############################################
#       check dmodel
#############################################


def _err(key, dconstraints, lvar):
    return (
        f"For model '{key}' dconstraints must be  dict of the form:\n"
        "\t'g0': {'ref': var0, vari: [c0, c1, c2], varj: [c0, c1, c2]},\n"
        "\t'g1': {'ref': varn, varh: [c0, c1, c2], vark: [c0, c1, c2]},\n\n"
        "Where [c0, c1, c2] are coefficients of a quadratic polynom of ref:\n"
        "\t\t vari = c0 + c1 * ref + c2 * ref**2\n\n"
        "Where the var are keys to existing model function:\n"
        f"\t- available keys: {sorted(lvar)}\n\n"
        "Each key can only be used once\n"
        f"Provided:\n{dconstraints}"
    )


def _check(
    coll=None,
    key=None,
    dconstraints=None,
):

    # -------------
    # extract variale groups
    # -------------

    lvar = coll.get_spectral_model_variables(
        key,
        returnas='all',
        concatenate=True,
    )['all']

    # -------------
    # trivial
    # -------------

    if dconstraints is None:
        dconstraints = {}

    # -------------
    # model
    # -------------

    c0 = (
        isinstance(dconstraints, dict)
        and all([
            isinstance(k0, str)
            and isinstance(v0, dict)
            and all([
                k1 in lvar + ['ref']
                and (
                    (
                        k1 == 'ref'
                        and v1 in lvar
                    )
                    or
                    (
                        k1 in lvar
                        and hasattr(v1, '__iter__')
                        and len(v1) == 3
                    )
                )
                for k1, v1 in v0.items()
            ])
            for k0, v0 in dconstraints.items()
        ])
    )

    if c0 is False:
        raise Exception(_err(key, dconstraints, lvar))

    # ------------
    # Format
    # ------------

    dconstraints = copy.deepcopy(dconstraints)
    for k0, v0 in dconstraints.items():
        for k1, v1 in v0.items():
            if k1 != 'ref':
                dconstraints[k0][k1] = np.atleast_1d(v1).astype(float).ravel()

    # --------------------------------
    # check each key is used once only
    # --------------------------------

    lin = list(itt.chain.from_iterable([
        [v0['ref']] + [k1 for k1 in v0.keys() if k1 != 'ref']
        for v0 in dconstraints.values()
    ]))

    linu = list(set(lin))
    if len(lin) > len(linu):
        raise Exception(_err(key, dconstraints, lvar))

    # safety double-check
    assert all([k0 in lvar for k0 in lin]), (lvar, lin)

    # ------------------
    # add missing keys
    # ------------------

    for k0 in set(lvar).difference(lin):
        dconstraints[k0] = {'ref': k0}

    return dconstraints, lvar


#############################################
#############################################
#       compute matrix
#############################################


def _compute_coefs_offset(
    dconstraints=None,
    lvar=None,
):

    # ----------------
    # get matrix shape
    # ----------------

    lr = [
        k0 for k0 in lvar
        if any([vg['ref'] == k0 for vg in dconstraints.values()])
    ]
    lg = [
        [
            kg for kg, vg in dconstraints.items()
            if vg['ref'] == k0
        ][0]
        for k0 in lr
    ]

    ng = len(dconstraints)
    nvar = len(lvar)

    c2 = np.zeros((nvar, ng), dtype=float)
    c1 = np.zeros((nvar, ng), dtype=float)
    c0 = np.zeros((nvar,), dtype=float)

    # --------------------------
    # populate with constraints
    # --------------------------

    for ii, gg in enumerate(lg):

        iref = lvar.index(dconstraints[gg]['ref'])
        c1[iref, ii] = 1.

        for k1 in set(dconstraints[gg].keys()).difference(['ref']):
            ind = lvar.index(k1)

            c0[ind] = dconstraints[gg][k1][0]
            c1[ind, ii] = dconstraints[gg][k1][1]
            c2[ind, ii] = dconstraints[gg][k1][2]

    # ----------------
    # store
    # ----------------

    dconstraints = {
        'groups': lg,
        'dconst': dconstraints,
        'c0': c0,
        'c1': c1,
        'c2': c2,
    }

    return dconstraints


# -*- coding: utf-8 -*-
"""
Created on Wed Jul 10 22:11:23 2024

@author: dvezinet
"""

import os


import numpy as np


from .._class02_SpectralFit import SpectralFit as Collection


# ############################################################
# ############################################################
#                DEFAULTS
# ############################################################


_PATH_HERE = os.path.dirname(__file__)
_PFE = os.path.join(_PATH_HERE, 'input', 'HXR_pulses.npz')


# ############################################################
# ############################################################
#                main
# ############################################################


def main(
    coll=None,
    pfe=_PFE,
    # spectral fits to be run
    lspect_fit=['exp', 'gauss', 'lognorm'],
    strict=True,
    verb=1,
):

    # --------------
    # initialize
    # --------------

    if coll is None:
        coll = Collection()

    # --------------
    # load file
    # --------------

    dout = dict(np.load(pfe, allow_pickle=True))

    # --------------
    # populate
    # --------------

    # -------------
    # ref

    coll.add_ref('npulse', size=dout['pulse'].size)
    coll.add_ref('nsamp', size=dout['data'].shape[0])

    # -------------
    # data

    # pulses
    coll.add_data(
        key='pulses',
        data=dout['pulse'],
        ref='npulse',
        units='index',
    )

    # indices
    coll.add_data(
        key='sample',
        data=np.arange(0, dout['data'].shape[0]),
        ref='nsamp',
        units='index',
    )

    # data
    coll.add_data(
        key='current',
        data=dout['data'],
        ref=('nsamp', 'npulse'),
        units='A',
    )

    # --------------
    # spectral models
    # --------------

    # spectral constrainst
    dconstraints = {
        'gbck': {
            'ref': 'bck_a0',
            'bck_a1': [0, 0, 0],
            'bck_a2': [0, 0, 0],
        },
    }

    # single exponential pulse
    coll.add_spectral_model(
        key='sm_exp',
        dmodel={
            'bck': 'poly',
            'pulse': 'pulse_exp',
        },
        dconstraints=dconstraints,
    )

    # single gaussian pulse
    coll.add_spectral_model(
        key='sm_gauss',
        dmodel={
            'bck': 'poly',
            'pulse': 'pulse_gauss',
        },
        dconstraints=dconstraints,
    )

    # single lognorm pulse
    coll.add_spectral_model(
        key='sm_lognorm',
        dmodel={
            'bck': 'poly',
            'pulse': 'lognorm',
        },
        dconstraints=dconstraints,
    )

    # --------------
    # define spectral fits
    # --------------

    # single exponential pulse
    for k0 in lspect_fit:
        coll.add_spectral_fit(
            key=f'sf_{k0}',
            key_model=f'sm_{k0}',
            key_data='current',
            key_sigma=None,
            key_lamb='sample',
            # params
            dparams=None,
            dvalid={
                'mask': None,
                'domain': None,
                'focus': None,
                'focus_logic': None,
            },
        )

    # ---------------------
    # compute spectral fits
    # ---------------------

    for k0 in lspect_fit:

        coll.compute_spectral_fit(
            key=f"sf_{k0}",
            strict=strict,
            verb=verb,
            timing=None,
        )

    return coll


# ############################################################
# ############################################################
#                __main__
# ############################################################


if __name__ == "__main__":
    coll = main()


"""
This module contains tests for tofu.geom in its structured version
"""


# Built-in
import itertools as itt


# Standard
import numpy as np


# local
from ._setup_teardown import setup_module0, teardown_module0
from .. import openadas


#######################################################
#
#     Setup and Teardown
#
#######################################################


def setup_module(module):
    setup_module0(module)


def teardown_module(module):
    teardown_module0(module)


#######################################################
#
#     Creating Ves objects and testing methods
#
#######################################################

# DOES NOT SEEM TO WORK ON GITUB ANYMORE
# class Test_openadas(object):

#     # ------------------------
#     #   setup and teardown
#     # ------------------------

#     @classmethod
#     def setup_class(cls):
#         pass

#     @classmethod
#     def setup_method(self):
#         pass

#     def teardown_method(self):
#         pass

#     @classmethod
#     def teardown_class(cls):
#         pass

#     # ------------------------
#     #   search online
#     # ------------------------

#     def test01_search_online_by_searchstr(self):
#         out = openadas.step01_search_online(
#             'ar+16 ADF15',
#             verb=True,
#             # verb=False,
#             returnas=False,
#         )
#         assert out is None

#         out = openadas.step01_search_online(
#             'ar+16 ADF11',
#             verb=False,
#             returnas=np.ndarray,
#         )
#         assert isinstance(out, np.ndarray)

#     def test02_search_online_by_wavelength(self):
#         # lret = [None, str, np.ndarray]
#         lret = [str]
#         # lverb = [True, False]
#         lverb = [False]
#         lelement = [None, 'ar', ['ar', 'w'], ('w',)]
#         # lcharge = [None, 14, [15, 16], (16,)]
#         lcharge = [[15, 16], (16,)]
#         lres = ['transition', 'file']
#         for comb in itt.product(lret, lverb, lelement, lcharge, lres):
#             out = openadas.step01_search_online_by_wavelengthA(
#                 lambmin=3.94,
#                 lambmax=4.,
#                 returnas=comb[0],
#                 verb=comb[1],
#                 element=comb[2],
#                 charge=comb[3],
#                 resolveby=comb[4],
#             )
#             assert out is comb[0] or isinstance(out, comb[0])

#     # ------------------------
#     #       download
#     # ------------------------

#     def test03_download(self):
#         out = openadas.step02_download(
#             filename='/adf15/pec40][ar/pec40][ar_ic][ar16.dat',
#             update=False,
#             verb=False,
#             returnas=False,
#         )
#         assert out is None

#         out = openadas.step02_download(
#             filename='/adf15/pec40][ar/pec40][ar_ic][ar16.dat',
#             update=True,
#             verb=False,
#             returnas=str,
#         )
#         assert isinstance(out, str)

#         out = openadas.step02_download_all(
#             lambmin=3.94,
#             lambmax=4,
#             element='Mo',
#             update=False,
#             verb=True,
#         )
#         assert out is None

#         lf = [
#             '/adf15/pec40][ar/pec40][ar_ic][ar16.dat',
#             '/adf11/scd74/scd74_ar.dat',
#             '/adf15/pec40][w/pec40][w_ls][w64.dat',
#             '/adf11/plt41/plt41_xe.dat',
#         ]
#         out = openadas.step02_download_all(
#             files=lf,
#             update=False,
#             verb=False,
#         )
#         assert out is None

#         out = openadas.step02_download_all(
#             searchstr='ar+16 ADF15',
#             include_partial=False,
#             update=False,
#             verb=False,
#         )
#         assert out is None

#     # ------------------------
#     #   readfiles
#     # ------------------------

#     def test04_readfiles(self):
#         out = openadas.step03_read(
#             '/adf15/pec40][ar/pec40][ar_ic][ar16.dat',
#         )
#         assert isinstance(out, dict)

#         out = openadas.step03_read(
#             'adf11/plt41/plt41_xe.dat',
#         )
#         assert isinstance(out, dict)

#         out = openadas.step03_read_all(
#             element='ar',
#             charge=16,
#             typ1='adf15',
#             lambmin=3.94,
#             lambmax=4.,
#         )
#         assert isinstance(out, dict)

#         out = openadas.step03_read_all(
#             element='ar',
#             charge=16,
#             typ1='adf11',
#             typ2='scd',
#         )
#         assert isinstance(out, dict)

#         out = openadas.step03_read_all(
#             element='ar',
#             charge=16,
#             typ1='adf11',
#             typ2='plt',
#         )
#         assert isinstance(out, dict)

#     # ------------------------
#     #   clear
#     # ------------------------

#     def test05_clear_downloads(self):
#         openadas.clear_downloads()


# -*- coding: utf-8 -*-
#! /usr/bin/python
"""
The openadas-compatibility module of spectrally

"""


import traceback


try:
    try:
        from spectrally.openadas._requests import *
        from spectrally.openadas._read_files import *
    except Exception:
        from ._requests import *
        from ._read_files import *

except Exception as err:
    msg = str(traceback.format_exc())
    msg += "\n\n\t=> optional sub-package spectrally.openadas not usable\n"
    raise Exception(msg)

del traceback



# -*- coding: utf-8 -*-
"""
Created on Fri Aug 30 08:04:47 2024

@author: dvezinet
"""


import matplotlib.pyplot as plt
import matplotlib.colors as mcolors
import datastock as ds


# ##################################################################
# ##################################################################
#                    main
# ##################################################################


def main(
    coll=None,
    keys=None,
    labels=None,
    colors=None,
):

    # ---------------
    # check inputs
    # ---------------

    keys, defcolor, colors = _check(
        coll=coll,
        keys=keys,
        labels=labels,
        colors=colors,
    )

    # ----------------
    # labels
    # ----------------

    if labels is None:

        dlabels = {k0: {'label': k0} for k0 in keys}

    elif isinstance(labels, str):

        wsl = coll._which_lines
        dlabels = {
            k0: {'label': coll.dobj[wsl].get(k0, {}).get(labels, k0)}
            for k0 in keys
        }

    else:

        dlabels = {
            k0: {'label': labels.get(k0, k0)}
            for k0 in keys
        }

    # ----------------
    # colors
    # ----------------

    if colors is None:

        for k0 in dlabels.keys():
            dlabels[k0]['color'] = defcolor

    elif isinstance(colors, str):

        if mcolors.is_color_like(colors):
            for k0 in dlabels.keys():
                dlabels[k0]['color'] = colors

        elif colors in ['sum', 'details']:
            pass

        else:
            wsl = coll._which_lines

            # get unique values
            lunique = set([
                coll.dobj[wsl].get(k0, {}).get(colors, k0)
                for k0 in dlabels.keys()
            ])

            # get color cycle and loop
            color_cycle = plt.rcParams['axes.prop_cycle'].by_key()['color']
            ncol = len(color_cycle)
            for ii, uu in enumerate(lunique):
                for k0 in dlabels.keys():
                    if coll.dobj[wsl].get(k0, {}).get(colors, k0) == uu:
                        dlabels[k0]['color'] = color_cycle[ii%ncol]

    else:

        for k0 in dlabels.keys():
            dlabels[k0]['color'] = colors.get(k0, defcolor)

    return dlabels


# ##################################################################
# ##################################################################
#                    check
# ##################################################################


def _check(
    coll=None,
    keys=None,
    labels=None,
    colors=None,
):

    # ----------------
    # keys
    # ----------------

    # which
    wsl = coll._which_lines
    lok_lines = list(coll.dobj.get(wsl, {}).keys())

    if hasattr(coll, '_which_model'):
        wsm = coll._which_model
        wsf = coll._which_fit
        lok_model = list(coll.dobj.get(wsm, {}).keys())
        lok_fit = list(coll.dobj.get(wsf, {}).keys())
    else:
        lok_model = []
        lok_fit = []

    # --------
    # deal with by case

    if isinstance(keys, str):

        if keys in lok_fit + lok_model:

            if keys in lok_fit:
                keym = coll.dobj[wsf][keys][wsm]
            else:
                keym = keys

            keys = [
                k0 for k0, v0 in coll.dobj[wsm][keym]['dmodel'].items()
                if v0['type'] not in ['poly', 'exp_lamb']
            ]
            lok = keys

        else:
            keys = [keys]
            lok = lok_lines

    else:
        lok = lok_lines

    # ----------
    # check

    keys = ds._generic_check._check_var_iter(
        keys, 'keys',
        types=(list, tuple),
        types_iter=str,
        allowed=lok,
    )

    # ----------------
    # labels
    # ----------------

    if isinstance(labels, str):

        if coll.dobj.get(wsl) is None:
            msg = (
                "Arg 'labels', if str, must be a spectral line parameter!\n"
                "But you apparently have no spectral lines in your Collection!"
                f"\nProvided: '{labels}'\n"
            )
            raise Exception(msg)

        lparam = coll.get_lparam(wsl)

        if labels not in lparam:
            msg = (
                "Arg 'labels', if str, must be a spectral line parameter!\n"
                "\t available: {lparam}\n"
                f"\t Provided: '{labels}'\n"
            )
            raise Exception(msg)

    elif isinstance(labels, dict):

        c0 = all([
            isinstance(k0, str)
            and k0 in keys
            and isinstance(v0, str)
            for k0, v0 in labels.items()
        ])
        if not c0:
            msg = (
                "Arg 'labels', if dict, must be of the form:\n"
                "\t- dict: {'key0': 'label0', 'key1': 'label1', ...}\n"
                f"Provided:\n{labels}\n"
            )
            raise Exception(msg)

    elif labels is not None:
        msg = (
            "Arg 'labels' must be either:\n"
            "\t- str: a valid spectral line parameter"
            "\t\t e.g: 'symbol', 'ion', ...\n"
            "\t- dict: {'key0': 'label0', 'key1': 'label1', ...}\n"
            f"Provided:\n{labels}\n"
        )
        raise Exception(msg)

    # ----------------
    # colors
    # ----------------

    # default
    defcolor = 'k'
    if colors is None:
        if coll.dobj.get(wsl) is None:
            colors = defcolor
        else:
            colors = 'ion'

    # case by case
    if isinstance(colors, str):

        if mcolors.is_color_like(colors):
            pass

        elif colors in ['sum', 'details']:
            pass

        else:

            if coll.dobj.get(wsl) is None:
                msg = (
                    "Arg 'colors', if str and not any of these:\n"
                    "\t- any color-like str\n"
                    "\t- 'sum'\n"
                    "\t- 'details'\n"
                    "Then must be a spectral line parameter!\n"
                    "But you apparently have no spectral lines in your Collection!"
                    f"\nProvided: '{colors}'\n"
                )
                raise Exception(msg)

            lparam = coll.get_lparam(wsl)

            if colors not in lparam:
                msg = (
                    "Arg 'colors', if str, must be a spectral line parameter!\n"
                    "\t available: {lparam}\n"
                    f"\t Provided: '{colors}'\n"
                )
                raise Exception(msg)

    elif isinstance(colors, dict):

        c0 = all([
            isinstance(k0, str)
            and k0 in keys
            and isinstance(v0, str)
            for k0, v0 in colors.items()
        ])
        if not c0:
            msg = (
                "Arg 'colors', if dict, must be of the form:\n"
                "\t- dict: {'key0': 'color0', 'key1': 'color1', ...}\n"
                f"Provided:\n{colors}\n"
            )
            raise Exception(msg)

    elif colors is not None:
        msg = (
            "Arg 'colors' must be either:\n"
            "\t- str: a valid color-like str, 'sum', 'details' "
            "or a valid spectral line parameter"
            "\t\t e.g: 'symbol', 'ion', ...\n"
            "\t- dict: {'key0': 'color0', 'key1': 'color1', ...}\n"
            f"Provided:\n{colors}\n"
        )
        raise Exception(msg)

    return keys, defcolor, colors




import numpy as np
import matplotlib.pyplot as plt
from matplotlib import gridspec
# import matplotlib as mpl
import matplotlib.transforms as transforms
import matplotlib.lines as mlines
# import matplotlib.colors as mcolors
import datastock as ds


_WINTIT = ""


# ################################################################
# ################################################################
#                       Spectral Lines
# ################################################################


def _check_axvline_inputs(
    param_x=None,
    param_txt=None,
    sortby=None,
    ymin=None, ymax=None,
    ls=None, lw=None, fontsize=None,
    side=None, fraction=None,
):

    # ------
    # param_x

    param_x = ds._generic_check._check_var(
        param_x, 'param_x',
        types=str,
        default='lamb0',
    )

    # ------
    # param_txt

    param_txt = ds._generic_check._check_var(
        param_txt, 'param_txt',
        types=str,
        default='symbol',
    )

    # ------
    # sortby

    sortby = ds._generic_check._check_var(
        sortby, 'sortby',
        types=str,
        default='ion',
        allowed=['ion', 'ION', 'source'],
    )

    # ---------
    # others

    if ymin is None:
        ymin = 0
    if ymax is None:
        ymax = 1
    if ls is None:
        ls = '-'
    if lw is None:
        lw = 1.
    if fontsize is None:
        fontsize = 9
    if side is None:
        side = 'right'
    if fraction is None:
        fraction = 0.75

    return (
        param_x, param_txt, sortby,
        ymin, ymax, ls, lw, fontsize, side, fraction,
    )


def _ax_axvline(
    ax=None, figsize=None, dmargin=None,
    quant=None, units=None, xlim=None,
    wintit=None, tit=None,
):

    if ax is None:

        if figsize is None:
            figsize = (9, 6)
        if dmargin is None:
            dmargin = {
                'left': 0.10, 'right': 0.90,
                'bottom': 0.10, 'top': 0.90,
                'hspace': 0.05, 'wspace': 0.05,
            }
        if wintit is None:
            wintit = _WINTIT
        if tit is None:
            tit = ''

        fig = plt.figure(figsize=figsize)
        fig.canvas.manager.set_window_title(wintit)
        fig.suptitle(tit, size=12, fontweight='bold')

        gs = gridspec.GridSpec(1, 1, **dmargin)
        ax = fig.add_subplot(gs[0, 0])

        ax.set_ylim(0, 1)
        ax.set_xlim(xlim)
        ax.set_xlabel('{} ({})'.format(quant, units))

    return ax


def plot_axvlines(
    din=None, key=None,
    sortby=None, dsize=None,
    param_x=None, param_txt=None,
    ax=None, ymin=None, ymax=None,
    ls=None, lw=None, fontsize=None,
    side=None, dcolor=None,
    fraction=None, units=None,
    figsize=None, dmargin=None,
    wintit=None, tit=None,
):

    # ------------
    # Check inputs

    (
        param_x, param_txt, sortby,
        ymin, ymax, ls, lw, fontsize, side, fraction,
    ) = _check_axvline_inputs(
        param_x=param_x,
        param_txt=param_txt,
        sortby=sortby,
        ymin=ymin, ymax=ymax,
        ls=ls, lw=lw,
        fontsize=fontsize,
        side=side,
        fraction=fraction,
    )

    # ------------
    # Prepare data

    unique = sorted(set([din[k0][sortby] for k0 in key]))
    ny = len(unique)
    dy = (ymax-ymin)/ny
    ly = [(ymin+ii*dy, ymin+(ii+1)*dy) for ii in range(ny)]
    xside = 1.01 if side == 'right' else -0.01
    ha = 'left' if side == 'right' else 'right'

    if dcolor is None:
        lcol = plt.rcParams['axes.prop_cycle'].by_key()['color']
        dcolor = {uu: lcol[ii % len(lcol)] for ii, uu in enumerate(unique)}

    # -----------------------------
    # sizes for scatter if relevant

    if dsize is not None:
        x, y = [], []
        colors = []
        sizes = []
        for ii, uu in enumerate(unique):
            lk = [
                k0 for k0 in key
                if din[k0][sortby] == uu and k0 in dsize.keys()
            ]
            if len(lk) > 0:
                x.append([din[k0][param_x] for k0 in lk])
                y.append([ly[ii][0]+fraction*dy/2. for k0 in lk])
                colors.append([dcolor[uu] for ii in range(len(lk))])
                sizes.append([dsize[k0] for k0 in lk])

        x = np.concatenate(x).ravel()
        y = np.concatenate(y).ravel()
        sizes = np.concatenate(sizes).ravel()
        colors = np.concatenate(colors).ravel()

    # ----------------
    # plot preparation

    lamb = [din[k0][param_x] for k0 in key]
    Dlamb = np.nanmax(lamb) - np.nanmin(lamb)
    xlim = [np.nanmin(lamb) - 0.05*Dlamb, np.nanmax(lamb) + 0.05*Dlamb]
    ax = _ax_axvline(
        ax=ax, figsize=figsize, dmargin=dmargin,
        quant=param_x, units=units, xlim=xlim,
        wintit=wintit, tit=tit,
    )

    blend = transforms.blended_transform_factory(
        ax.transAxes, ax.transData
    )

    # ----
    # plot

    for ii, uu in enumerate(unique):
        lk = [k0 for k0 in key if din[k0][sortby] == uu]
        for k0 in lk:
            _ = ax.axvline(
                x=din[k0][param_x],
                ymin=ly[ii][0],
                ymax=ly[ii][0] + fraction*dy,
                c=dcolor[uu],
                ls=ls,
                lw=lw,
            )

            ax.text(
                din[k0][param_x],
                ly[ii][1],
                din[k0][param_txt],
                color=dcolor[uu],
                horizontalalignment='center',
                verticalalignment='top',
                fontsize=fontsize,
                fontweight='normal',
                transform=ax.transData,
            )
        ax.text(
            xside,
            0.5*(ly[ii][0] + ly[ii][1]),
            uu,
            color=dcolor[uu],
            horizontalalignment=ha,
            verticalalignment='center',
            fontsize=fontsize+1,
            fontweight='bold',
            transform=blend,
        )

    # -----------------------------------
    # Add scatter plot if dsizes provided

    if dsize is not None:
        ax.scatter(
            x, y, s=sizes**2, c=colors,
            marker='o', edgecolors='None',
        )

    return ax


# #############################################################################
# #############################################################################
#               Dominance map
# #############################################################################


# def _dominance_map_check(

# ):

    # if param_txt is None:
        # param_txt = 'symbol'
    # if param_color is None:
        # param_color = 'ion'
    # if norder is None:
        # norder = 0

    # if ne_scale is None:
        # ne_scale = 'log'
    # if Te_scale is None:
        # Te_scale = 'poly'

    # return

"""
def _ax_dominance_map(
    dax=None, figsize=None, dmargin=None,
    x_scale=None, y_scale=None, amp_scale=None,
    quant=None, units=None,
    wintit=None, tit=None, dtit=None,
    proj=None, dlabel=None,
):

    # -------------
    # Check inputs

    pass

    # -------------
    # Prepare data

    # Get dcolor
    lcol = plt.rcParams['axes.prop_cycle'].by_key()['color']
    dcolor = {}
    if param_color != 'key':
        lion = [self._dobj['lines'][k0][param_color] for k0 in dpec.keys()]
        for ii, k0 in enumerate(set(lion)):
            dcolor[k0] = mcolors.to_rgb(lcol[ii % len(lcol)])
            lk1 = [
                k2 for k2 in dpec.keys()
                if self._dobj['lines'][k2][param_color] == k0
            ]
            for k1 in lk1:
                damp[k1]['color'] = k0
    else:
        for ii, k0 in enumerate(dpec.keys()):
            dcolor[k0] = mcolors.to_rgb(lcol[ii % len(lcoil)])
            damp[k0]['color'] = k0

    # Create image
    im_data = np.full((ne_grid.size, Te_grid.size), np.nan)
    im = np.full((ne_grid.size, Te_grid.size, 4), np.nan)
    dom_val = np.concatenate(
        [v0[None, :, :] for v0 in dpec_grid.values()],
        axis=0,
    )

    if norder == 0:
        im_ind = np.nanargmax(dom_val, axis=0)
    else:
        im_ind = np.argsort(dom_val, axis=0)[-norder, :, :]

    for ii in np.unique(im_ind):
        ind = im_ind == ii
        im_data[ind] = dom_val[ii, ind]

    pmin = np.nanmin(np.log10(im_data))
    pmax = np.nanmax(np.log10(im_data))

    for ii, k0 in enumerate(dpec_grid.keys()):
        if ii in np.unique(im_ind):
            ind = im_ind == ii
            im[ind, :-1] = dcolor[damp[k0]['color']]
            im[ind, -1] = (
                (np.log10(im_data[ind])-pmin)/(pmax-pmin)*0.9 + 0.1
            )
    extent = (ne_grid.min(), ne_grid.max(), Te_grid.min(), Te_grid.max())

    if tit is None:
        tit = 'spectral lines PEC interpolations'
    if dtit is None:
        dtit = {'map': 'norder = {}'.format(norder)}

    # ------------
    # prepare axes

    allowed = ['map', 'spect', 'amp', 'prof']
    if dax is None:

        proj = _check_proj(proj=proj, allowed=allowed)
        dax = {}

        if figsize is None:
            figsize = (15, 9)
        if dmargin is None:
            dmargin = {
                'left': 0.05, 'right': 0.90,
                'bottom': 0.08, 'top': 0.90,
                'hspace': 0.20, 'wspace': 0.50,
            }
        if wintit is None:
            wintit = _WINTIT
        if tit is None:
            tit = ''
        if dtit is None:
            dtit = {}

        fig = plt.figure(figsize=figsize)
        fig.canvas.manager.set_window_title(wintit)
        fig.suptitle(tit, size=12, fontweight='bold')

        if len(proj) == 1:
            gs = gridspec.GridSpec(1, 1, **dmargin)
            for k0 in allowed:
                dax[k0] = fig.add_subplot(gs[0, 0])
        else:
            gs = gridspec.GridSpec(3, 5, **dmargin)
            shx, shy = None, None
            k0 = 'map'
            if k0 in proj:
                dax[k0] = fig.add_subplot(
                    gs[:, :2], xscale=x_scale, yscale=y_scale,
                )
            if 'spect' in proj:
                dax['spect'] = fig.add_subplot(
                    gs[0, 2:], yscale=amp_scale,
                )
                shy = dax['spect']
            if 'amp' in proj:
                dax['amp'] = fig.add_subplot(
                    gs[1, 2:], sharey=shy, yscale=amp_scale,
                )
                shx = dax['amp']
            if 'prof' in proj:
                dax['prof'] = fig.add_subplot(
                    gs[2, 2:], yscale=x_scale, sharex=shx,
                )

        for k0 in proj:
            if dtit is not None and dtit.get(k0) is not None:
                dax[k0].set_title(dtit[k0])

        for k0 in proj:
            if dlabel is not None and dlabel.get(k0) is not None:
                dax[k0].set_xlabel(dlabel[k0]['x'])
                dax[k0].set_ylabel(dlabel[k0]['y'])

    else:
        c0 = (
            isinstance(dax, dict)
            and all([ss in allowed for ss in dax.keys()])
        )
        if not c0:
            msg = (
                "\nArg dax must be a dict with the following allowed keys:\n"
                + "\t- allowed:  {}\n".format(allowed)
                + "\t- provided: {}".format(sorted(dax.keys()))
            )
            raise Exception(msg)

    return dax


def plot_dominance_map(
    din=None, key=None,
    im=None, extent=None,
    xval=None, yval=None, damp=None,
    x_scale=None, y_scale=None, amp_scale=None,
    sortby=None, dsize=None,
    param_x=None, param_txt=None,
    dax=None, proj=None,
    ls=None, lw=None, fontsize=None,
    side=None, dcolor=None,
    fraction=None, units=None,
    figsize=None, dmargin=None,
    wintit=None, tit=None, dtit=None, dlabel=None,
):

    # Check inputs

    # Prepare dax
    dax = _ax_dominance_map(
        dax=dax, proj=proj, figsize=figsize, dmargin=dmargin,
        quant=param_x, units=units,
        wintit=wintit, tit=tit, dtit=dtit, dlabel=dlabel,
        x_scale=x_scale, y_scale=y_scale, amp_scale=amp_scale,
    )

    if any([ss in dax.keys() for ss in ['spect', 'prof', 'amp']]):
        ind = np.arange(0, xval.size)

    k0 = 'map'
    if dax.get(k0) is not None:
        dax[k0].imshow(
            np.swapaxes(im, 0, 1),
            extent=extent,
            origin='lower',
            aspect='auto',
        )
        dax[k0].plot(xval, yval, ls='-', marker='.', lw=1., c='k')

    k0 = 'prof'
    if dax.get(k0) is not None:
        dax[k0].plot(ind, xval, ls='-', marker='.', lw=1., c='k')
        dax[k0].plot(ind, yval, ls='-', marker='.', lw=1., c='k')

    k0 = 'amp'
    if dax.get(k0) is not None:
        for k1 in damp.keys():
            dax[k0].plot(
                ind, damp[k1]['data'],
                ls='-', marker='.', lw=1., c=dcolor[damp[k1]['color']],
            )

    k0 = 'spect'
    if dax.get(k0) is not None:
        lamb = np.ones((list(damp.values())[0]['data'].size,))
        blend = transforms.blended_transform_factory(
            dax[k0].transData, dax[k0].transAxes
        )
        for k1 in damp.keys():
            lamb0 = din[k1]['lamb0']
            dax[k0].plot(
                lamb0*lamb, damp[k1]['data'],
                ls='None', marker='.', lw=1., c=dcolor[damp[k1]['color']],
            )
            dax[k0].text(
                lamb0,
                1.,
                din[k1][param_txt],
                color=dcolor[damp[k1]['color']],
                horizontalalignment='left',
                verticalalignment='bottom',
                fontsize=fontsize,
                fontweight='normal',
                transform=blend,
                rotation=60,
            )

        handles = [
            mlines.Line2D([], [], color=v0, label=k0)
            for k0, v0 in dcolor.items()
        ]
        dax[k0].legend(
            handles=handles,
            loc=2, bbox_to_anchor=(1., 1.),
        )

    return dax
"""


# #!/usr/bin/python3
# -*- coding: utf-8 -*-


import os
import json
import itertools as itt


import numpy as np
import datastock as ds


#############################################
#############################################
#       add
#############################################


def add_lines(
    coll=None,
    key=None,
    ion=None,
    lamb0=None,
    transition=None,
    pec=None,
    source=None,
    symbol=None,
    **kwdargs,
):

    # -----------
    # check

    ion, lamb0, transition, source, symbol = _check(
        coll=coll,
        key=key,
        ion=ion,
        lamb0=lamb0,
        transition=transition,
        pec=pec,
        source=source,
        symbol=symbol,
    )

    # -----------
    # add ion

    if ion is not None and ion not in coll.dobj.get('ion', {}).keys():
        coll.add_ion(ion)

    # --------
    # add line

    coll.add_obj(
        which=coll._which_lines,
        key=key,
        ion=ion,
        lamb0=lamb0,
        transition=transition,
        pec=pec,
        source=source,
        symbol=symbol,
        **kwdargs,
    )

    return


#############################################
#############################################
#       check
#############################################


def _check(
    coll=None,
    key=None,
    ion=None,
    lamb0=None,
    transition=None,
    pec=None,
    source=None,
    symbol=None,
):

    # -----------------
    # check consistency
    # -----------------




    # ---------------
    # check items
    # ---------------


    # ------------
    # ion

    if ion is not None:
        ion = ds._generic_check._check_var(
            ion, 'ion',
            types=str,
        )

    # ------------
    # lamb0

    lamb0 = float(ds._generic_check._check_var(
        lamb0, 'lamb0',
        types=(int, float),
        sign=">0",
    ))

    # ------------
    # transition

    if transition is not None:
        transition = ds._generic_check._check_var(
            transition, 'transition',
            types=str,
        )

    # ------------
    # source

    if source is not None:
        source = ds._generic_check._check_var(
            source, 'source',
            types=str,
        )

    # ------------
    # symbol

    symbol = ds._generic_check._obj_key(
        {
            v0.get('symbol'): None
            for v0 in coll.dobj.get(coll._which_lines, {}).values()
        },
        short='l',
        key=symbol,
        ndigits=3,
    )

    return ion, lamb0, transition, source, symbol


#############################################
#############################################
#       add ions from lines dobj
#############################################


def _add_lines_from_dobj(coll=None, dref=None, ddata=None, dobj=None):

    # ---------
    # add ions

    lions = sorted(dobj.get('ion', {}).keys())
    if len(lions) > 0:
        for ion in lions:
            coll.add_ion(ion)
        del dobj['ion']

    # -------
    # update

    coll.update(ddata=ddata, dref=dref, dobj=dobj)

    return


#############################################
#############################################
#      remove lines
#############################################


def remove_lines(
    coll=None,
    keys=None,
    propagate=None,
):

    # ----------------
    # check inputs
    # ----------------

    # --------
    # keys

    if isinstance(keys, str):
        keys = [keys]

    wsl = coll._which_lines
    lok = list(coll.dobj.get(wsl, {}).keys())
    keys = ds._generic_check._check_var_iter(
        keys, 'keys',
        types=(list, tuple),
        types_iter=str,
        allowed=lok,
        default=lok,
    )

    # --------------
    # propagate

    propagate = ds._generic_check._check_var(
        propagate, 'propagate',
        types=bool,
        default=True,
    )

    # -----------------
    # propagate
    # -----------------

    if propagate is True:

        dlines = coll.dobj.get(wsl)
        lout = [k0 for k0 in dlines.keys() if k0 not in keys]

        # ions
        lion_in = set([
            dlines[k0]['ion'] for k0 in keys
            if dlines[k0].get('ion') is not None
        ])
        lion_out = set([
            dlines[k0]['ion'] for k0 in lout
            if dlines[k0].get('ion') is not None
        ])
        lions = list(lion_in.difference(lion_out))

        # sources
        ls_in = set([
            dlines[k0]['source'] for k0 in keys
            if dlines[k0].get('source') is not None
        ])
        ls_out = set([
            dlines[k0]['source'] for k0 in lout
            if dlines[k0].get('source') is not None
        ])
        lsources = list(ls_in.difference(ls_out))

        # pec
        lpec_in = set([
            dlines[k0]['pec'] for k0 in keys
            if dlines[k0].get('pec') is not None
        ])
        lpec_out = set([
            dlines[k0]['pec'] for k0 in lout
            if dlines[k0].get('pec') is not None
        ])
        lpec = list(lpec_in.difference(lpec_out))

        # lref
        lref = list(set(itt.chain.from_iterable([
            coll.ddata[k0]['ref'] for k0 in lpec
        ])))

    # -----------------
    # remove
    # -----------------

    coll.remove_obj(which=coll._which_lines, key=keys)

    if propagate is True:
        if len(lions) > 0:
            coll.remove_obj(which='ion', key=lions)
        if len(lref) > 0:
            coll.remove_ref(lref, propagate=True)
        if len(lpec) > 0:
            coll.remove_data(lpec, propagate=True)
        if len(lsources) > 0:
            coll.remove_obj(which='source', key=lsources)

    return


#############################################
#############################################
#       save lines to file
#############################################


def _save_to_file(
    coll=None,
    keys=None,
    path=None,
    name=None,
    overwrite=None,
):

    # ---------------
    # check inputs
    # ---------------

    keys, pfe, overwrite = _check_to_file(
        coll=coll,
        keys=keys,
        path=path,
        name=name,
        overwrite=overwrite,
    )

    # --------------
    # extract lines
    # ---------------

    lparam = ['ion', 'source', 'lamb0', 'symbol', 'type', 'transition']
    dlines = coll.dobj.get(coll._which_lines, {})
    dout = {
        k0: {k1: dlines[k0][k1] for k1 in lparam}
        for k0 in keys
    }

    # --------------
    # save to file
    # --------------

    if pfe.endswith('json'):

        with open(pfe, 'w+') as fn:
            json.dump(
                dout,
                fn,
                indent=4,
                sort_keys=True,
                ensure_ascii=False,
                check_circular=True,
                allow_nan=True,
                separators=(', ', ': '),
            )

    else:
        np.savez(pfe, **dout)

    # ---------------
    # verb

    msg = f"Saved to:\n\t{pfe}"
    print(msg)

    return


#############################################
#############################################
#       save lines to file - check
#############################################


def _check_to_file(coll=None, keys=None, path=None, name=None, overwrite=None):

    # ---------------
    # check keys
    # ---------------

    if isinstance(keys, str):
        keys = [keys]

    lok = coll.dobj.get(coll._which_lines, {}).keys()
    keys = ds._generic_check._check_var_iter(
        keys, 'keys',
        types=(list, tuple),
        types_iter=str,
        default=sorted(lok),
        allowed=lok,
    )

    # ---------------
    # check inputs
    # ---------------

    if path is None:
        path = '.'

    if not (isinstance(path, str) and os.path.isdir(path)):
        msg = (
            "Arg path must be a str, a valid path to a directory!\n"
            f"Provided:\n\t{path}\n"
        )
        raise Exception(msg)

    path = os.path.abspath(path)

    # ---------------
    # check inputs
    # ---------------

    name = ds._generic_check._check_var(
        name, 'name',
        types=str,
        default='spectrallines',
    )

    if name.count('.') == 0:
        name = f'{name}.json'

    if not any([name.endswith(ss) for ss in ['.npz', '.json']]):
        msg = (
            "Arg name must a file name ending in '.json' or '.npz'!\n"
            f"Provided:\n\t{name}\n"
        )
        raise Exception(msg)

    pfe = os.path.join(path, name)

    # ---------------
    # check inputs
    # ---------------

    overwrite = ds._generic_check._check_var(
        overwrite, 'overwrite',
        types=bool,
        default=False,
    )

    return keys, pfe, overwrite



# #!/usr/bin/python3
# -*- coding: utf-8 -*-


import copy


from ._class00_SpectralLines import SpectralLines as Previous
from . import _class01_check_model as _check_model
from . import _class01_show as _show
from . import _class01_check_constraints as _check_constraints
from . import _class01_fit_func as _fit_func
from . import _class01_interpolate as _interpolate
from . import _class01_moments as _moments
from . import _class01_plot as _plot


__all__ = ['SpectralModel']


#############################################
#############################################
#       Spectral Lines
#############################################


class SpectralModel(Previous):

    _which_model = 'spect_model'
    _which_fit = 'spect_fit'

    _ddef = copy.deepcopy(Previous._ddef)

    # _show_in_summary_core = ['shape', 'ref', 'group']
    _dshow = dict(Previous._dshow)
    _dshow[_which_model] = ['']

    _ddef['params']['dobj'] = {
    }

    # ###################
    # -------------------
    # Spectral models
    # -------------------

    # -------------------
    # add spectral model
    # -------------------

    def add_spectral_model(
        self,
        key=None,
        dmodel=None,
        dconstraints=None,
    ):
        """ Add a spectral model for future fitting

        Defined by a set of functions and constraints.
            - dmodel: dict of (key, function type) pairs
            - dconstraints: dict of:
                'key': {'ref': k0, k1: [c0, c1], k2: [c0, c1]}

        Available function types for dmodel are:
            - 'poly': typically a polynomial background
            - 'exp': typically an exponential background
            - 'gauss': typically a doppler-broadened line
            - 'lorentz': ypically a natural-broadened line
            - 'pvoigt': typically a doppler-and-natural broadened line

        dconstraints holds a dict of constraints groups.
        Each group is a dict with a 'ref' variable
        Other variables (keys) are compued as poly functions of 'ref'

        Parameters
        ----------
        key : str, optional
            DESCRIPTION. The default is None.
        dmodel : dict, optional
            DESCRIPTION. The default is None.
        dconstraints : dict, optional
            DESCRIPTION. The default is None.

        """

        # check and store the model
        _check_model._dmodel(
            coll=self,
            key=key,
            dmodel=dmodel,
        )

        # check and store the constraints
        _check_constraints._dconstraints(
            coll=self,
            key=key,
            dconstraints=dconstraints,
        )

    # -------------------
    # show spectral model
    # -------------------

    def _get_show_obj(self, which=None):
        if which == self._which_model:
            return _show._show
        else:
            return super()._get_show_obj(which)

    def _get_show_details(self, which=None):
        if which == self._which_model:
            return _show._show_details
        else:
            super()._get_show_details(which)

    # -------------------
    # get spectral model
    # -------------------

    def get_spectral_model_variables(
        self,
        key=None,
        returnas=None,
        concatenate=None,
    ):
        """ Get ordered list of individual variable names """

        return _check_model._get_var(
            coll=self,
            key=key,
            returnas=returnas,
            concatenate=concatenate,
        )

    def get_spectral_model_variables_dind(
        self,
        key=None,
    ):
        """ Get ordered list of individual variable names """

        return _check_model._get_var_dind(
            coll=self,
            key=key,
        )

    # ----------------------
    # interpolate spectral model
    # ----------------------

    def interpolate_spectral_model(
        self,
        key_model=None,
        key_data=None,
        lamb=None,
        # options
        details=None,
        binning=None,
        # uncertainty propagation
        uncertainty_method=None,
        # others
        returnas=None,
        store=None,
        store_key=None,
        # timing
        timing=None,
    ):
        """ Interpolate the spectral model at lamb using key_data


        Parameters
        ----------
        key_model : str, optional
            key to the desired spectral model
        key_data : str, optional
            key to the data to be used for the model's free variables
                - has to have the model's ref in its own references
        lamb : str/np.ndarray
            DESCRIPTION. The default is None.

        Returns
        -------
        dout : dict
            output dict of interpolated data, with units and ref

        """

        return _interpolate.main(
            coll=self,
            key_model=key_model,
            key_data=key_data,
            lamb=lamb,
            # options
            details=details,
            binning=binning,
            # uncertainty propagation
            uncertainty_method=uncertainty_method,
            # others
            returnas=returnas,
            store=store,
            store_key=store_key,
            # timing
            timing=timing,
        )

    # ----------------------
    # extract spectral model moments
    # ----------------------

    def get_spectral_model_moments(
        self,
        key=None,
        key_data=None,
        lamb=None,
        returnas=None,
    ):
        """

        Given an array of solution for the model's free variables (key_data)
        Returns, a dict with, for each function in the model:

            - all variable values (claasified by type)
            - some moments of interest:
                - integral
                - physics-derived (Te from exp, Ti from line width...)

        Parameters
        ----------
        key_model : str
            key to the desired spectral model (or spectral fit)
        key_data : str
            key to the data to be used as input for the model's free variables
        lamb:str or 1d np.ndarray, optional
            wavelenth vector to be used for computing limited integrals
        returnas: str
            Flag indicating whether to return dout as:
                - 'dict_arrays': nested dict of {'ktype': {'kvar': array}}
                - 'dict_varnames': dict of {'kfunc_var': values}

        Returns
        -------
        dout : dict
            dict of all moments, classified per type and function

        """

        dout = _moments.main(
            coll=self,
            key=key,
            key_data=key_data,
            lamb=lamb,
            returnas=returnas,
        )

        return dout

    # ###################
    # -------------------
    # Spectral fits
    # -------------------

    # -------------------
    # get func details, cost, jac
    # -------------------

    def get_spectral_fit_func(
        self,
        key=None,
        func=None,
    ):
        """ Return the fitting functions for a given model
        """

        return _fit_func.main(
            coll=self,
            key=key,
            func=func,
        )

    # ----------------------------
    # plot spectral modela and fit
    # ----------------------------

    def plot_spectral_model(
        self,
        key_model=None,
        key_data=None,
        lamb=None,
        keyY=None,
        # options
        details=None,
        # plotting
        dprop=None,
        vmin=None,
        vmax=None,
        # figure
        dax=None,
        fs=None,
        dmargin=None,
        tit=None,
    ):
        """ Plot a spectral model using specified data

        lamb can be:
            - a key to an existing vector
            - a user-provided vector (1d np.ndarray)

        """

        return _plot.main(
            coll=self,
            key_model=key_model,
            key_data=key_data,
            lamb=lamb,
            keyY=keyY,
            # options
            details=details,
            # plotting
            dprop=dprop,
            vmin=vmin,
            vmax=vmax,
            # figure
            dax=dax,
            fs=fs,
            dmargin=dmargin,
            tit=tit,
        )



# Built-in
import os
import shutil
import requests
import warnings

# Common
import numpy as np


__all__ = [
    'step01_search_online',
    'step01_search_online_by_wavelengthA',
    'step02_download',
    'step02_download_all',
    'clear_downloads',
]


# Check whether a local .spectrally/ repo exists
_URL = 'https://open.adas.ac.uk'
_URL_SEARCH = _URL + '/freeform?searchstring='
_URL_SEARCH_WAVL = _URL + '/wavelength?'
_URL_ADF15 = _URL + '/adf15'
_URL_DOWNLOAD = _URL + '/download'
_CUSTOM = os.path.dirname(os.path.dirname(os.path.dirname(__file__)))
_CUSTOM = os.path.join(_CUSTOM, 'scripts', 'spectrallycustom.py')

_CREATE_CUSTOM = True
_INCLUDE_PARTIAL = True


_DCERTIFICATES_BUNDLE = {
    'ITER': {
        'host': 'iter.org',
        'bund': '/etc/pki/ca-trust/extracted/openssl/ca-bundle.trust.crt',
    },
}


# #############################################################################
#                           Utility functions
# #############################################################################


def _get_PATH_LOCAL():
    pfe = os.path.join(os.path.expanduser('~'), '.spectrally', 'openadas')
    if os.path.isdir(pfe):
        return pfe
    else:
        return None


def _getcharray(ar, col=None, sep='  ', line='-', just='l',
                verb=True, returnas=str):
    """ Format and return char array (for pretty printing) """

    # Trivial case
    c0 = ar is None or len(ar) == 0
    if c0:
        if returnas is str:
            return ''
        elif returnas is np.ndarray:
            return np.array([])
        elif returnas is False:
            return

    # Non-trivial cases
    ar = np.array(ar, dtype='U')

    if ar.ndim == 1:
        ar = ar.reshape((1, ar.size))

    # Get just len
    nn = np.char.str_len(ar).max(axis=0)
    if col is not None:
        if len(col) not in ar.shape:
            msg = ("len(col) should be in np.array(ar, dtype='U').shape:\n"
                   + "\t- len(col) = {}\n".format(len(col))
                   + "\t- ar.shape = {}".format(ar.shape))
            raise Exception(msg)
        if len(col) != ar.shape[1]:
            ar = ar.T
            nn = np.char.str_len(ar).max(axis=0)
        nn = np.fmax(nn, [len(cc) for cc in col])

    # Apply to array
    fjust = np.char.ljust if just == 'l' else np.char.rjust
    out = np.array([sep.join(v) for v in fjust(ar, nn)])

    # Apply to col
    if col is not None:
        arcol = np.array([col, [line*n for n in nn]], dtype='U')
        arcol = np.array([sep.join(v) for v in fjust(arcol, nn)])
        out = np.append(arcol, out)

    if verb is True:
        print('\n'.join(out))

    if returnas is str:
        return '\n'.join(out)
    elif returnas is np.ndarray:
        return ar
    elif returnas is False:
        return


# #############################################################################
#                          online search - utilities
# #############################################################################


class FileAlreayExistsException(Exception):
    pass


def _try_request_handle_ITER_SSL_step01(url=None):
    try:
        resp = requests.get(url)
    except requests.exceptions.SSLError as err:
        if 'certificate' in str(err):
            lbund = [
                vv['bund'] for kk, vv in _DCERTIFICATES_BUNDLE.items()
                if os.uname()[1].endswith(vv['host'])
            ]
            if len(lbund) == 1:
                resp = requests.get(url, verify=lbund[0])
            else:
                msg = (
                    str(err)
                    + "\n\nLooks like a certificate error occured!\n"
                    + "=> try changing the certificate bundle of requests\n"
                    + "=> ask your admin which certificate bundle to use!"
                )
                raise Exception(msg)
        else:
            raise err
    except Exception as err:
        raise err
    return resp


# #############################################################################
#                          online search
# #############################################################################


def step01_search_online(
    searchstr=None, returnas=None,
    include_partial=None, verb=None,
):
    """ Perform an online freeform search on https://open.adas.ac.uk

    Pass searchstr to the online freeform search
    Prints the results (if verb=True)

    Optionally return the result as:
        - np.ndarray    : a char array
        - str           : a formatted str

    example
    -------
        >>> import spectrally as tf
        >>> tf.openadas.step01_search_online('ar+16 ADF15')
    """

    # Check input
    if returnas is None:
        returnas = False
    if verb is None:
        verb = True
    if include_partial is None:
        include_partial = _INCLUDE_PARTIAL
    if searchstr is None:
        searchstr = ''
    searchurl = '+'.join([requests.utils.quote(kk)
                          for kk in searchstr.split(' ')])

    total_url = '{}{}&{}'.format(_URL_SEARCH, searchurl, 'searching=1')
    resp = _try_request_handle_ITER_SSL_step01(url=total_url)

    # Extract response from html
    out = resp.text.split('\n')
    flag0 = '<table summary="Freeform search results">'
    flag1 = '</table>'
    ind0 = [ii for ii, vv in enumerate(out) if flag0 in vv]
    ind1 = [ii for ii, vv in enumerate(out) if flag1 in vv]
    if len(ind0) != 1 or len(ind1) == 0:
        msg = ("Format of html response seems to have changed!\n"
               + "Cannot find flags:\n"
               + "\t- {}\n".format(flag0)
               + "\t- {}\n".format(flag1)
               + "in requests.get({}).text".format(total_url))
        raise Exception(msg)
    ind1 = np.min([ii for ii in ind1 if ii > ind0[0]])
    out = out[ind0[0] + 1:ind1-1]
    nresults = len(out) - 1

    # Get columns
    heads = [str.replace(kk.replace('<tr>', '').replace('<th>', ''),
                         '</th>', '').replace('</tr>', '')
             for kk in out[0].split('</th><th>')]
    nhead = len(heads)

    # Get results
    lout = []
    for ii in range(0, nresults):
        if 'Partial results are listed below' in out[ii+1]:
            if include_partial is False:
                break
            lout.append(['-', '-', '----- (partial results) -----', '-'])
            ind = out[ii+1].index('below</th></tr>') + len('below</th></tr>')
            out[ii+1] = out[ii+1][ind:]
        lstri = out[ii+1].split('</td><td')
        assert len(lstri) == nhead
        elmq = lstri[0].replace('<tr><td>', '')
        if '<sup>' in elmq:
            elm = elmq[:elmq.index('<sup>')]
            charge = elmq[elmq.index('<sup>')+len('<sup>'):]
            charge = charge.replace('</sup>', '')
            if charge == '+':
                charge = '1+'
        else:
            charge = ''
            elm = elmq
        typ = lstri[2][1:]
        fil = lstri[3][lstri[3].index('detail')+len('detail'):]
        fil = fil[:fil.index('.dat')+len('.dat')]
        lout.append([elm, charge, typ, fil])

    # Format output
    char = np.array(lout)
    col = ['Element', 'charge', 'type of data', 'full file name']
    arr = _getcharray(char, col=col,
                      sep='  ', line='-', just='l',
                      returnas=returnas, verb=verb)
    return arr


def step01_search_online_by_wavelengthA(
    lambmin=None, lambmax=None,
    element=None, charge=None, resolveby=None,
    returnas=None, verb=None,
):
    """ Perform an online search by wavelength on https://open.adas.ac.uk

    Pass the min / max wavelength (in Angstrom) to the online wavelength search
    Prints the results (if verb=True)

    Optionally return the result as:
        - np.ndarray    : a char array
        - str           : a formatted str

    The result can be resolved by:
        - 'transition'  : by spectral transition
        - 'file'        : by adas file

    Optionally filter by element and charge

    example
    -------
        >>> import spectrally as tf
        >>> tf.openadas.step01_earch_online_by_wavelengthA(3., 4., 'ar')
        >>> tf.openadas.step01_earch_online_by_wavelengthA(
            3., 4., 'w',
            resolveby='file',
        )
    """

    # -----------
    # Check input

    # general
    if returnas is None:
        returnas = False
    if verb is None:
        verb = True
    if lambmin is None:
        lambmin = ''
    if lambmax is None:
        lambmax = ''
    if resolveby is None:
        resolveby = 'transition'
    if resolveby not in ['transition', 'file']:
        msg = ("Arg resolveby must be:\n"
               + "\t- 'transition': list all available transitions\n"
               + "\t- 'file': list all files containing relevant transitions")
        raise Exception(msg)

    # element
    if element is not None:
        c0 = (
            isinstance(element, str)
            or (
                isinstance(element, list)
                and all([isinstance(ee, str) for ee in element])
            )
            or (
                isinstance(element, tuple)
                and all([isinstance(ee, str) for ee in element])
            )
        )
        if not c0:
            msg = ("Arg element must be a str (e.g.: element='ar')\n"
                   + "\t- provided: {}".format(element))
            raise Exception(msg)
        if isinstance(element, str):
            element = [element]
        if isinstance(element, list):
            element = [ee.lower() for ee in element]
        elif isinstance(element, tuple):
            element = tuple([ee.lower() for ee in element])

    # charge
    if charge is not None:
        c0 = (
            isinstance(charge, int)
            or (
                isinstance(charge, list)
                and all([isinstance(cc, int) for cc in charge])
            )
            or (
                isinstance(charge, tuple)
                and all([isinstance(cc, int) for cc in charge])
            )
        )
        if not c0:
            msg = ("Arg charge must be a int or list (e.g.: 16 or [0])\n"
                   + "\t- provided: {}".format(charge))
            raise Exception(msg)
        if isinstance(charge, int):
            charge = [charge]
        if isinstance(charge, list):
            charge = ['0' if cc == 0 else '{}+'.format(cc) for cc in charge]
        elif isinstance(charge, tuple):
            charge = tuple([
                '0' if cc == 0 else '{}+'.format(cc) for cc in charge
            ])

    # ---------------
    # prepare request

    searchurl = '&'.join(['wave_min={}'.format(lambmin),
                          'wave_max={}'.format(lambmax),
                          'resolveby={}'.format(resolveby)])

    total_url = '{}{}&{}'.format(_URL_SEARCH_WAVL, searchurl, 'searching=1')
    resp = _try_request_handle_ITER_SSL_step01(url=total_url)

    # Extract response from html
    out = resp.text.split('\n')
    flag0 = '<table summary="Search by Wavelength Search Results">'
    flag1 = '</table></div></div></div>'
    ind0 = [ii for ii, vv in enumerate(out) if flag0 in vv]
    ind1 = [ii for ii, vv in enumerate(out) if flag1 in vv]
    if len(ind0) != 1 or len(ind1) != 1:
        msg = ("Format of html response seems to have changed!\n"
               + "Cannot find flags:\n"
               + "\t- {}\n".format(flag0)
               + "\t- {}\n".format(flag1)
               + "in requests.get({}).text".format(total_url))
        raise Exception(msg)
    out = out[ind0[0] + 1].split('</tr><tr><td>')
    nresults = len(out) - 1

    # Get columns
    col = [kk.replace('<tr><th>', '').replace('</th>', '').strip()
           for kk in out[0].split('</th><th>')]
    ncol = len(col)
    if resolveby == 'transition':
        dcolex = {
            'Wavelength': 0,
            'Ion': 1,
            'Data Type': 2,
            'Transition': 3,
            'File Details': 4,
        }
    else:
        dcolex = {
            'Ion': 0,
            'Data Type': 1,
            'Minimum Wavelength': 2,
            'Maximum Wavelength': 3,
            'File Details': 4,
        }
    if col != list(dcolex.keys()):
        msg = (
            "Format of table columns in html seems to have changed!\n"
            f"\t- expected: {dcolex}\n"
            f"\t- observed: {col}\n"
        )
        raise Exception(msg)

    lout = []
    if resolveby == 'transition':
        for ii in range(0, nresults):
            lstri = out[ii+1].split('</td><td>')
            assert len(lstri) == ncol
            elm, charg = (
                lstri[dcolex['Ion']].replace('</sup>', '').split('<sup>')
            )
            if charg == '+':
                charg = '1+'
            c0 = (
                (isinstance(element, list) and elm.lower() not in element)
                or (isinstance(element, tuple) and elm.lower() in element)
            )
            if c0:
                continue
            c0 = (
                (isinstance(charge, list) and charg not in charge)
                or (isinstance(charge, tuple) and charg in charge)
            )
            if c0:
                continue
            lamb = lstri[dcolex['Wavelength']].replace('&Aring;', '')
            typ = (
                lstri[dcolex['Data Type']].replace('</span>', '').split('>')[1]
            )
            trans = lstri[dcolex['Transition']].replace('&nbsp;', ' ')
            trans = trans.replace('<sup>', '^{').replace('</sup>', '}')
            trans = trans.replace('<sub>', '_{').replace('</sub>', '}')
            trans = trans.replace('&rarr;', '->')
            fil = lstri[dcolex['File Details']]
            fil = fil[fil.index('detail')+len('detail'):]
            fil = fil[:fil.index('.dat')+len('.dat')]
            lout.append([lamb, elm, charg, typ, trans, fil])
    else:
        for ii in range(0, nresults):
            lstri = out[ii+1].split('</td><td>')
            elm, charg = (
                lstri[dcolex['Ion']].replace('</sup>', '').split('<sup>')
            )
            if charg == '+':
                charg = '1+'
            c0 = (
                (isinstance(element, list) and elm.lower() not in element)
                or (isinstance(element, tuple) and elm.lower() in element)
            )
            if c0:
                continue
            c0 = (
                (isinstance(charge, list) and charg not in charge)
                or (isinstance(charge, tuple) and charg in charge)
            )
            if c0:
                continue
            typ = (
                lstri[dcolex['Data Type']].replace('</span>', '')
            )
            lambmin = (
                lstri[dcolex['Minimum Wavelength']].replace('&Aring;', '')
            )
            lambmax = (
                lstri[dcolex['Maximum Wavelength']].replace('&Aring;', '')
            )
            fil = lstri[dcolex['File Details']]
            fil = fil[fil.index('detail')+len('detail'):]
            fil = fil[:fil.index('.dat')+len('.dat')]
            lout.append([fil, lambmin, lambmax, elm, charg, typ])

    # Format output
    char = np.array(lout)
    if resolveby == 'transition':
        col = ['Wavelength', 'Element', 'Charge', 'Data Type',
               'Transition', 'Full file name']
    else:
        col = ['Full file name', 'Wavelength min', 'Wavelength max',
               'Element', 'Charge', 'Data Type']
    arr = _getcharray(char, col=col,
                      sep='  ', line='-', just='l',
                      returnas=returnas, verb=verb)
    return arr


# #############################################################################
#                          Download - utilities
# #############################################################################


def _try_request_handle_ITER_SSL_step02(url=None, pfe=None):
    try:
        with requests.get(url, stream=True) as rr:
            rr.raise_for_status()
            with open(pfe, 'wb') as ff:
                for chunk in rr.iter_content(chunk_size=8192):
                    # filter-out keep-alive new chunks
                    if chunk:
                        ff.write(chunk)
                        # ff.flush()
    except requests.exceptions.SSLError as err:
        if 'certificate' in str(err):
            lbund = [
                vv['bund'] for kk, vv in _DCERTIFICATES_BUNDLE.items()
                if os.uname()[1].endswith(vv['host'])
            ]
            if len(lbund) == 1:
                with requests.get(url, stream=True, verify=lbund[0]) as rr:
                    rr.raise_for_status()
                    with open(pfe, 'wb') as ff:
                        for chunk in rr.iter_content(chunk_size=8192):
                            # filter-out keep-alive new chunks
                            if chunk:
                                ff.write(chunk)
                                # ff.flush()
            else:
                msg = (
                    str(err)
                    + "\n\nLooks like a certificate error occured!\n"
                    + "=> try changing the certificate bundle of requests\n"
                    + "=> ask your admin which certificate bundle to use!"
                )
                raise Exception(msg)
        else:
            raise err
    except Exception as err:
        raise err


def _check_exists(filename, update=None, create_custom=None):

    # In case a small modification becomes necessary later
    target = filename
    path_local = _get_PATH_LOCAL()

    # Check whether the local .spectrally repo exists
    # if not recommend spectrally-custom
    if path_local is None:
        path = os.path.join(os.path.expanduser('~'), '.spectrally', 'openadas')
        if create_custom is None:
            create_custom = _CREATE_CUSTOM
        if create_custom is True:
            os.system('python ' + _CUSTOM)
            path_local = _get_PATH_LOCAL()

        if path_local is None or create_custom is False:
            msg = (
                "You do not seem to have a local .spectrally repository\n"
                "spectrally uses that local repository to store user-specific "
                "data and downloads\n"
                "In particular, openadas files are downloaded in:\n"
                + f"\t{path}\n"
                "  => to set-up your local .spectrally repo, run in terminal:\n"
                "\tspectrally custom"
            )
            raise Exception(msg)

    # Parse intermediate repos and create if necessary
    lrep = target.split('/')[1:-1]
    for ii in range(len(lrep)):
        repo = os.path.join(path_local, *lrep[:ii+1])
        if not os.path.isdir(repo):
            os.mkdir(repo)

    # Check if file already exists
    path = os.path.join(path_local, *lrep)
    pfe = os.path.join(path, target.split('/')[-1])
    if os.path.isfile(pfe):
        if update is False:
            msg = ("File already exists in your local repo:\n"
                   + "\t{}\n".format(pfe)
                   + "  => if you want to force download, use update=True"
                   + " (local file will be overwritten)")
            raise FileAlreayExistsException(msg)
        else:
            return True, pfe
    else:
        return False, pfe


# #############################################################################
#                          Download
# #############################################################################


def step02_download(
    filename=None,
    update=None,
    create_custom=None,
    verb=None,
    returnas=None,
):
    """ Download desired file from  https://open.adas.ac.uk

    All downloaded files are stored in your local spectrally directory
    (~/.spectrally/)

    Automatically runs spectrally-custom if create_custom=True

    example
    -------
        >>> import spectrally as sp
        >>> filename = '/adf15/pec40][ar/pec40][ar_ls][ar16.dat'
        >>> sp.openadas.step02_download(filename)
    """

    # ---------------------------
    # Check
    if verb is None:
        verb = True
    if update is None:
        update = False
    if returnas is None:
        returnas = False

    c0 = (not isinstance(filename, str)
          or filename[:4] != '/adf' or filename[-4:] != '.dat')
    if c0:
        msg = ("filename must be a str (full file name) of the form:\n"
               + "\t/adf.../.../....dat\n"
               + "\nProvided:\n\t{}".format(filename))
        raise Exception(msg)
    url = _URL_DOWNLOAD + filename

    exists, pfe = _check_exists(
        filename, update=update,
        create_custom=create_custom
    )
    if exists is True and verb is True:
        msg = ("File already exists, will be downloaded and overwritten:\n"
               + "\t{}".format(pfe))
        warnings.warn(msg)

    # ---------------------------
    # Download
    # Note the stream=True parameter below
    _try_request_handle_ITER_SSL_step02(url=url, pfe=pfe)

    if verb is True:
        msg = ("file {} was copied to:\n".format(filename)
               + "\t{}".format(pfe))
        print(msg)
    if returnas is str:
        return pfe


def step02_download_all(
    files=None, searchstr=None,
    lambmin=None, lambmax=None, element=None,
    include_partial=None, update=None, create_custom=None, verb=None,
):
    """ Download all desired files from  https://open.adas.ac.uk

    The files to download can be provided either as:
        - a list of full openadas file names (files)
        - the result of an online freeform search
            (searchstr fed to search_online)
        - the input to an online search by wavelength
            (lambmin, lambmax and element fed to search_online_by_wavelengthA)

    Automatically runs spectrally-custom if create_custom=True
    All downloaded files are stored in your local spectrally directory
    (~/.spectrally/)

    example
    -------
        >>> import spectrally as sp
        >>> sp.openadas.step02_download_all(
            lambmin=3., lambmax=4., element='ar',
        )
    """

    # Check
    if include_partial is None:
        include_partial = False
    if update is None:
        update = False
    if verb is None:
        verb = True
    lc = [files is not None,
          searchstr is not None,
          any([ss is not None for ss in [lambmin, lambmax, element]])]
    if np.sum(lc) != 1:
        msg = (
            "Please either searchstr xor (lambmin, lambmax, element)\n"
            + "\t- files: list of full file names to be downloaded\n"
            + "\t- searchstr: uses step01_search_online()\n"
            + "\t- lambmin, lambmax, element: "
            + "uses step01_search_online_by_wavelengthA()")
        raise Exception(msg)

    # Get list of files
    if lc[0]:
        if isinstance(files, str):
            files = [files]
        if not (isinstance(files, list)
                and all([isinstance(ss, str) for ss in files])):
            msg = "files must be a list of full openadas file names!"
            raise Exception(msg)
    elif lc[1]:
        arr = step01_search_online(
            searchstr=searchstr,
            include_partial=include_partial,
            verb=False,
            returnas=np.ndarray,
        )
        files = arr[:, -1]
    elif lc[2]:
        arr = step01_search_online_by_wavelengthA(
            lambmin=lambmin,
            lambmax=lambmax,
            element=element, verb=False,
            returnas=np.ndarray,
        )
        files = np.unique(arr[:, -1])

    # Download
    if verb is True:
        msg = "Downloading from {} into {}:".format(_URL, _get_PATH_LOCAL())
        print(msg)
    for ii in range(len(files)):
        try:
            exists, pfe = _check_exists(
                files[ii], update=update,
                create_custom=create_custom,
            )
            pfe = step02_download(
                filename=files[ii], update=update,
                verb=False, returnas=str,
            )
            if exists is True:
                msg = "\toverwritten:   \t{}".format(files[ii])
            else:
                msg = "\tdownloaded:    \t{}".format(files[ii])
        except FileAlreayExistsException:
            msg = "\talready exists:  {}".format(files[ii])
        except Exception as err:
            msg = (str(err)
                   + "\n\nCould not download file {}".format(files[ii]))
            raise err

        if verb is True:
            print(msg)


def clear_downloads():
    """ Delete all openadas files downloaded in your ~/.spectrally/ directory """
    path_local = _get_PATH_LOCAL()
    if path_local is None:
        return
    lf = [
        os.path.join(path_local, ff)
        for ff in os.listdir(path_local)
        if os.path.isfile(os.path.join(path_local, ff))
    ]
    ld = [
        os.path.join(path_local, ff) for ff in os.listdir(path_local)
        if os.path.isdir(os.path.join(path_local, ff))
    ]
    for ff in lf:
        os.remove(ff)
    for dd in ld:
        shutil.rmtree(dd)


# -*- coding: utf-8 -*-
"""
Created on Wed Jul 24 09:32:47 2024

@author: dvezinet
"""


# common
import numpy as np
import matplotlib.pyplot as plt
from matplotlib import gridspec
import datastock as ds


# local
from . import _class01_model_dict as _model_dict


__all__ = ['display_spectral_model_function']


# #######################################################
# #######################################################
#            Main
# #######################################################


def display_spectral_model_function(
    # func type
    ftype=None,
    # plot vector
    xx=None,
    binning=None,
    # plotting
    dax=None,
    # resources
    dmodel=_model_dict._DMODEL,
):

    plt.rcParams['text.usetex'] = True

    # -------------
    # check inputs
    # -------------

    ftype = ds._generic_check._check_var(
        ftype, 'ftype',
        types=str,
        allowed=list(dmodel.keys()),
    )

    dfunc = dmodel[ftype]

    # --------------
    # prepare data
    # --------------

    # get plot vector
    xx = _get_xx(xx, ftype)

    # get dparam and free variables from plot vector
    dpar, xfree = _get_dpar_xfree(ftype, xx)

    # -----------------
    # prepare Collection
    # -----------------

    # initialize
    from ._class02_SpectralFit import SpectralFit as Collection
    coll = Collection()

    # add model
    coll.add_spectral_model(
        key='sm00',
        dmodel={
            'func': {'type': ftype, **dpar},
        },
        dconstraints=None,
    )

    # add ref
    coll.add_ref('n0', size=xx.size)

    # add data
    coll.add_data('x', data=xx)

    # add xfree
    coll.add_data('var', data=xfree)

    # --------------
    # compute model
    # --------------

    coll.interpolate_spectral_model(
        key_model='sm00',
        key_data='var',
        lamb='x',
        store=True,
        store_key='y',
    )

    # --------------
    # prepare figure
    # --------------

    if dax is None:
        dax = _create_dax(
            binning=binning,
            fs=None,
            dmargin=None,
            tit=None,
        )

    dax = ds._generic_check._check_dax(dax)

    # --------------
    # plot
    # --------------

    _plot(
        coll=coll,
        ftype=ftype,
        dfunc=dfunc,
        dax=dax,
    )

    return coll, dax


# #######################################################
# #######################################################
#            Prepare x, y
# #######################################################


def _get_xx(xx, ftype):

    # -------------
    # set if None
    # -------------

    lbck = ['poly', 'exp_lamb']
    llines = ['gauss', 'lorentz', 'pvoigt', 'voigt']
    lpulse = ['pulse_exp', 'pulse_gauss', 'lognorm']

    if xx is None:
        if ftype in lbck + llines:
            xx = np.linspace(3.94, 4, 501) * 1e-10

        elif ftype in lpulse:
            xx = np.linspace(100, 300, 501) * 1e-9

        else:
            raise NotImplementedError()

    # -------------
    # check
    # -------------

    try:
        xx = np.unique(np.atleast_1d(xx).ravel().astype(float))
    except Exception as err:
        msg = "Arg xx must be a 1d increasing array of floats"
        raise Exception(msg) from err

    return xx


# #######################################################
# #######################################################
#            Get dpar and xfree
# #######################################################


def _get_dpar_xfree(ftype, xx):

    # ---------------
    # prepare from xx
    # ---------------

    x0 = xx[0]
    x1 = xx[-1]
    dx = xx[1] - x0
    Dx = xx[-1] - x0
    xm = np.mean(xx)
    xmin = np.min(xx)
    xmax = np.max(xx)

    # ---------------
    # dpar
    # ---------------

    llines = ['gauss', 'lorentz', 'pvoigt', 'voigt']
    if ftype in llines:
        dpar = {'lamb0': xm}
    else:
        dpar = {}

    # ---------------
    # xfree
    # ---------------

    if ftype == 'poly':

        a0 = 1
        a1 = -0.2
        a2 = -1

        xfree = np.r_[a0, a1, a2]

    elif ftype == 'exp_lamb':

        rate = np.nanmax([
            np.abs(np.log(xmax * x0 / (xmin * x1)) / (1/x1 - 1./x0)),
            np.abs(np.log(xmin * x0 / (xmax * x1)) / (1/x1 - 1./x0)),
        ])
        amp = 1 * np.exp(rate/xm) * xm

        xfree = np.r_[amp, rate]

    elif ftype == 'gauss':

        amp = 1
        sigma = Dx / 10
        vccos = 0.

        xfree = np.r_[amp, vccos, sigma]

    elif ftype == 'lorentz':

        amp = 1
        gam = Dx / 10
        vccos = 0.

        xfree = np.r_[amp, vccos, gam]

    elif ftype == 'pvoigt':

        amp = 1
        sigma = Dx / 10
        gam = Dx / 10
        vccos = 0.

        xfree = np.r_[amp, vccos, sigma, gam]

    elif ftype == 'voigt':

        amp = 1
        sigma = Dx / 10
        gam = Dx / 10
        vccos = 0.

        xfree = np.r_[amp, vccos, sigma, gam]

    elif ftype == 'pulse_exp':

        amp = 1
        t0 = 0.3
        t_up = Dx / 30
        t_down = Dx / 5

        xfree = np.r_[amp, t0, t_up, t_down]

    elif ftype == 'pulse_gauss':

        amp = 1
        t0 = 0.3
        t_up = Dx / 30
        t_down = Dx / 5

        xfree = np.r_[amp, t0, t_up, t_down]

    elif ftype == 'lognorm':

        t0 = 0.3
        sigma = 1
        std = Dx / 5
        mu = 0.5 * (np.log(std**2/(np.exp(sigma**2) - 1)) - sigma**2)
        amp = (xmax - xmin) * np.exp(mu - 0.5*sigma**2)

        xfree = np.r_[amp, t0, mu, sigma]

    return dpar, xfree


# #######################################################
# #######################################################
#            plot
# #######################################################


def _plot(
    coll=None,
    ftype=None,
    dfunc=None,
    dax=None,
):

    # --------------
    # plot data
    # --------------

    axtype = 'plot'
    lax = [kax for kax, vax in dax.items() if axtype in vax['type']]
    for kax in lax:
        ax = dax[kax]['handle']

        ax.plot(
            coll.ddata['x']['data'],
            coll.ddata['y']['data'],
            ls='-',
            marker='None',
            lw=1,
            c='k',
        )

    # ----------------
    # plot expression
    # ----------------

    axtype = 'text'
    lax = [kax for kax, vax in dax.items() if axtype in vax['type']]
    for kax in lax:
        ax = dax[kax]['handle']

        # main expression
        ax.text(
            0.,
            0.5,
            dfunc['expressions']['main'],
            size=28,
            fontweight='bold',
            transform=ax.transAxes,
            horizontalalignment='left',
            verticalalignment='center',
        )

        # reference
        ref = dfunc['expressions'].get('ref')
        if ref is not None:
            ax.text(
                0.,
                0.1,
                ref,
                size=16,
                fontweight='bold',
                transform=ax.transAxes,
                horizontalalignment='left',
                verticalalignment='center',
            )

        # others
        nexp = len(dfunc['expressions']) - 1
        for ii, (k0, v0) in enumerate(dfunc['expressions'].items()):

            if k0 in ['main', 'ref']:
                continue

            ax.text(
                0.6,
                ii / (nexp+1),
                v0,
                size=18,
                fontweight='bold',
                transform=ax.transAxes,
                horizontalalignment='left',
                verticalalignment='center',
            )

    # ----------------
    # Figure title
    # ----------------

    ax.figure.suptitle(ftype, size=18, fontweight='bold')

    return


# #######################################################
# #######################################################
#            Figure
# #######################################################


def _create_dax(
    binning=None,
    fs=None,
    dmargin=None,
    tit=None,
):

    # ---------------
    # check inputs
    # ---------------

    if fs is None:
        fs = (14, 8)

    if dmargin is None:
        dmargin = {
            'left': 0.07, 'right': 0.98,
            'bottom': 0.08, 'top': 0.90,
            'wspace': 0.50, 'hspace': 0.10,
        }

    # ---------------
    # prepare figure
    # ---------------

    dax = {}
    fig = plt.figure(figsize=fs)
    gs = gridspec.GridSpec(2, 1, **dmargin)

    # ------------
    # add axes
    # ------------

    # plot
    ax = fig.add_subplot(gs[0, 0], frameon=True)
    # ax.set_xlabel()
    # ax.set_ylabel()
    dax['plot'] = {'handle': ax, 'type': 'plot'}

    # text
    ax = fig.add_subplot(gs[1, 0], sharex=ax, frameon=False)
    ax.set_xlabel('')
    ax.set_ylabel('')
    ax.set_yticks([])
    plt.setp(ax.get_xticklabels(), visible=False)
    dax['text'] = {'handle': ax, 'type': 'text'}

    return dax